#include <windows.h>
#include <math.h>
#include <assert.h>
#include <alloc.h>
//#include <stdlib.h>
#include "ca.hpp"
#include "random.h"
#include "camenu.h"

#define MIN(x,y) ((x)<=(y))?(x):(y)

#define BLT_LINES 3 //4 is faster, but I don't like the way it looks
	//Used to chunk lines to put on screen.
#define STATERADIUSCOUNT 10
int StartStates[STATERADIUSCOUNT] = {2,4,8,16,2,4,2,4,2,2};
int StartRadius[STATERADIUSCOUNT] = {1,1,1, 1,2,2,3,3,4,5};
static int blt_flag = 0; // this is for the timing of bitblting to screen

CA::CA(CAlist *mylist)
{ //enter unallocated, exit allocated, initialized.
	calist_ptr = mylist;
	/*Set this stuff before Allocate */
	hwnd = calist_ptr->hwnd;
	sourcerowindex = 0;
	targetrowindex = 1;
	pastrowindex = MEMORY-1;
	wavesourceindex = 0;
	wavetargetindex = 1;
	generatorflag = START_GENERATOR_FLAG;
	/*End of stuff needed before Allocate*/
	Allocate();
	InitSeed();	// Set all cells in all rows to zero
	/*Set this before you do any table settings */
	lambda = START_LAMBDA;
	oldradius = START_RADIUS;
	oldstates = START_STATES;
	nabeoptions = 0;
	cellcount = 0;
	radius = 0; //So that Changeradiusandstates does something
	states = 0;
	/*End of stuff needed for table settings */
	/* Put this before you do a Seed */
	showmode = START_SHOWMODE;
	time_step = START_MIN_TIME_STEP + (Randomreal() *
		START_MAX_TIME_STEP);
	space_step = START_MIN_SPACE_STEP + (Randomreal() *
		START_MAX_SPACE_STEP);
	time = 0.0;
	friction_multiplier = Randomreal() *
		START_MAX_FRICTION_MULTIPLIER;
	spring_multiplier = START_MIN_SPRING + Randomreal()*START_MAX_SPRING;
	mass = START_MIN_MASS + (Randomreal() *
		START_MAX_MASS);
	frequency_multiplier = 1 + Randomreal() *
		START_MAX_FREQUENCY_MULTIPLIER;
	max_intensity = START_MIN_INTENSITY +
		Randomreal() * (START_MAX_INTENSITY - START_MIN_INTENSITY);
	driver_multiplier = (max_intensity / 4.0) + Randomreal() *
		(max_intensity / 2.0);
	max_velocity = START_MIN_VELOCITY +
		Randomreal() * (START_MAX_VELOCITY - START_MIN_VELOCITY);
	band_count = START_BAND_COUNT;
	heat_inc = START_HEAT_INC;
	Adjust_acceleration_multiplier();
   //sets half_time_step and accleration_multiplier and frequency_factor
	/* End of stuff needed for seed */
	Changeradiusandstates(START_RADIUS, START_STATES);
	/*Do after Allocate, because it calls Lambdalookup, Seed, Resetfreq,
		Setcolortable */
	type_ca = 0; // Use 0 for a no-good type_ca.
	Settype(START_TYPE_CA); /*Won't do a Changeradiusandstates as oldstates
		is not over 16*/
	wrapflag = START_WRAPFLAG;
	entropy = 0.0;
	target_entropy = START_TARGET_ENTROPY;
	entropyflag = START_ENTROPYFLAG;
	entropy_bonus = START_ENTROPY_BONUS;
	fail_stripe = START_FAILSTRIPE_SCORE;
	score = 0.0;
};

void CA::Allocate()
{
	generator_ptr = new Generator(this);
	lookup = (unsigned char far *)(new char[MAXNABEOPTIONS]);
	if (!lookup)
	{
		MessageBox( hwnd,
			(LPSTR)"Failure in Lookup Allocation!",
			(LPSTR)"Memory Problems!",
			MB_OK | MB_ICONEXCLAMATION );
		SendMessage( hwnd, WM_DESTROY, 0, 0L );
		return;
	}
	freqlookup = (unsigned int far *)(new int[MAXNABEOPTIONS]);
	if (!freqlookup)
	{
		MessageBox( hwnd,
			(LPSTR)"Failure in Freqlookup Allocation!",
			(LPSTR)"Memory Problems!",
			MB_OK | MB_ICONEXCLAMATION );
		SendMessage( hwnd, WM_DESTROY, 0, 0L );
		return;
	}
	for (int i=0; i<MEMORY; i++)
	{
		rowbuffer[i] = (unsigned char far *) new char[
			GetSystemMetrics(SM_CXSCREEN)];
		if (!rowbuffer[i])
		{
			MessageBox( hwnd,
				(LPSTR)"Failure in Rowbuffer Allocation!",
				(LPSTR)"Memory Problems!",
					MB_OK | MB_ICONEXCLAMATION );
				SendMessage( hwnd, WM_DESTROY, 0, 0L );
				return;
		}
	}
	source_row = rowbuffer[sourcerowindex];
	target_row = rowbuffer[targetrowindex];
	past_row = rowbuffer[pastrowindex];
	for (i = 0; i < 2; i++)
	{
		waverowbuffer[i] = new Wavecell[
			GetSystemMetrics(SM_CXSCREEN)];
		if (!waverowbuffer[i])
		{
			MessageBox( hwnd,
				(LPSTR)"Failure in Waverow Allocation!",
				(LPSTR)"Memory Problems!",
					MB_OK | MB_ICONEXCLAMATION );
				SendMessage( hwnd, WM_DESTROY, 0, 0L );
				return;
		}
	}
	wave_source_row = waverowbuffer[wavesourceindex];
	wave_target_row = waverowbuffer[wavetargetindex];
}

CA::~CA()
{
	int i;

	delete lookup;
	delete freqlookup;
	for (i = 0; i<MEMORY; i++)
		delete rowbuffer[i];
	for (i = 0; i < 2; i++)
		delete waverowbuffer[i];
}

void CA::Lambdalookup()
{
	for (unsigned short i = 0; i < nabeoptions; i++)
	{
		if (Randomreal() < lambda)
			lookup[i] = 1 + Random(states - 1);
		else
			lookup[i] = 0;
	}
//---CA_wave
		time_step = START_MIN_TIME_STEP + Randomreal() *
			START_MAX_TIME_STEP;
		space_step = START_MIN_SPACE_STEP + Randomreal() *
			START_MAX_SPACE_STEP;
		mass = START_MIN_MASS + Randomreal() *
			START_MAX_MASS;
		friction_multiplier = Randomreal() *
			START_MAX_FRICTION_MULTIPLIER;
		spring_multiplier = START_MIN_SPRING + (Randomreal() *
			START_MAX_SPRING);
		frequency_multiplier = 1+ Randomreal() *
			START_MAX_FREQUENCY_MULTIPLIER;
		max_intensity = START_MIN_INTENSITY +
			Randomreal() * (START_MAX_INTENSITY - START_MIN_INTENSITY);
		driver_multiplier = (max_intensity / 4.0) + Randomreal() *
			(max_intensity / 2.0);
		max_velocity = START_MIN_VELOCITY +
			Randomreal() * (START_MAX_VELOCITY - START_MIN_VELOCITY);
		Adjust_acceleration_multiplier();

	Resetfreq();
	Computeactual_lambda();
	if (calist_ptr->FocusCA() == this)
		update_flag = 1;	// need to update dialog boxs
}

void CA::Resetfreq()
{
	cellcount = 0;
	for (unsigned short i = 0; i < nabeoptions; i++)
			freqlookup[i] = 0;
}

void CA::InitSeed()
{
	for (short j = 0; j < MEMORY; j++)
		for (short i = 0; i < GetSystemMetrics(SM_CXSCREEN); i++)
			rowbuffer[j][i] = 0;
	for (short i = 0; i < GetSystemMetrics(SM_CXSCREEN); i++)
	{
		wave_source_row[i].intensity = 0.0;
		wave_source_row[i].velocity = 0.0;
		wave_target_row[i].intensity = 0.0;
		wave_target_row[i].velocity = 0.0;
	}
}

void CA::Seed()
{
	// Reset all the rows to the top
	source_row = rowbuffer[0];
	sourcerowindex = 0;
	target_row = rowbuffer[1];
	targetrowindex = 1;
	past_row = rowbuffer[MEMORY - 1];
	pastrowindex = MEMORY - 1;
	wave_source_row = waverowbuffer[0];
	wavesourceindex = 0;
	wave_target_row = waverowbuffer[1];
	wavetargetindex = 1;
	for (int i = 0; i < GetSystemMetrics(SM_CXSCREEN); i++)
	{
		source_row[i] = Random(states);
		past_row[i] = Random(states);
		wave_source_row[i].intensity = 0.5 * Randomsignreal() * max_intensity;
		wave_source_row[i].velocity = 0.0;//Randomsignreal() * max_velocity;
	/* Need to put consistent colorstates pattern into source_row
	for meaningful Avoidstripes check in wave cases */
		if (type_ca != CA_STANDARD && type_ca != CA_REVERSIBLE)
		{
			if (!(calist_ptr->showvelocity))
			source_row[i] = (int) ((MAX_COLOR *
				(wave_source_row[i].intensity +
				max_intensity)) / (2.0 * max_intensity));
			else
			source_row[i] = (int) ((MAX_COLOR *
				(wave_source_row[i].velocity +
				max_velocity)) / (2.0 * max_velocity));
		}
	}
	Resetfreq();
	generator_ptr->Seed();
}

void CA::Sineseed()
{
	// Reset all the rows to the top
	source_row = rowbuffer[0];
	sourcerowindex = 0;
	target_row = rowbuffer[1];
	targetrowindex = 1;
	past_row = rowbuffer[MEMORY - 1];
	pastrowindex = MEMORY - 1;
	wave_source_row = waverowbuffer[0];
	wavesourceindex = 0;
	wave_target_row = waverowbuffer[1];
	wavetargetindex = 1;
	for (int i = 0; i < GetSystemMetrics(SM_CXSCREEN); i++)
	{
/* Let's set it for two complete sinewave cycles to fit into the
size of the CA window at seed time.  So I want to cover 4*PI in
the horz_count steps)*/
		source_row[i] = (unsigned char)(((Real)states/2.0) +
			sin((i * 4.0 * PI)/(horz_count)) * ((Real)states/2.0));
		if (source_row[i]>states-1)
			source_row[i] = states-1;
		past_row[i] = (unsigned char)(((Real)states/2.0) +
			sin((i * 4.0 * PI)/(horz_count)) * ((Real)states/2.0));
		if (past_row[i]>states-1)
			past_row[i] = states-1;
		wave_source_row[i].intensity = 0.5*sin((i * 4.0 * PI)/(horz_count)) *
			max_intensity;
/* The velocity should be the derivative d intensity /dx, where x
is just i * (4.0*PI/WAVE_PERIOD).  This is
as follows.  I don't think it would be right to have the derivative
be in terms of i with a chain rule term of (2.0*PI/WAVE_PERIOD).  Note
if max_velocity is less than max_intensity, this will get clipped soon.
But it sucks to have the velocities get clipped right away, so let's
use max_velocity after all.  Let's try 0.0*/
		wave_source_row[i].velocity = 0.0;
		 //	cos((i * 4.0 * PI)/(horz_count)) * max_velocity;
	/* Need to put consistent colorstates pattern into source_row
	for meaningful Avoidstripes check in wave cases */
		if (type_ca != CA_STANDARD && type_ca != CA_REVERSIBLE)
		{
			if (!(calist_ptr->showvelocity))
			source_row[i] = (int) ((MAX_COLOR *
				(wave_source_row[i].intensity +
				max_intensity)) / (2.0 * max_intensity));
			else
			source_row[i] = (int) ((MAX_COLOR *
				(wave_source_row[i].velocity +
				max_velocity)) / (2.0 * max_velocity));
		}

	}
	Resetfreq();
	generator_ptr->Seed();
}

void CA::Oneseed()
{
	// Reset all the rows
	source_row = rowbuffer[0];
	sourcerowindex = 0;
	target_row = rowbuffer[1];
	targetrowindex = 1;
	past_row = rowbuffer[MEMORY - 1];
	pastrowindex = MEMORY - 1;
	wave_source_row = waverowbuffer[0];
	wavesourceindex = 0;
	wave_target_row = waverowbuffer[1];
	wavetargetindex = 1;
	for (int i = 0; i < GetSystemMetrics(SM_CXSCREEN); i++)
	{
		source_row[i] = 0;
		past_row[i] = 0;
		wave_source_row[i].intensity = 0.0;
		wave_source_row[i].velocity = 0.0;
		if (type_ca != CA_STANDARD && type_ca != CA_REVERSIBLE)
			source_row[i] = (int) ((MAX_COLOR *
				(wave_source_row[i].intensity +
				max_intensity)) / (2.0 * max_intensity));

	}
	source_row[horz_count/2] = 1;
	wave_source_row[horz_count/2].intensity = max_intensity;
	//leave velocity zero.
	//wave_source_row[horz_count/2].velocity = max_velocity;
	/* Need to put consistent colorstates pattern into source_row
	for meaningful Avoidstripes check in wave cases */
		if (type_ca != CA_STANDARD && type_ca != CA_REVERSIBLE)
			source_row[horz_count/2] = (int) ((MAX_COLOR *
				(wave_source_row[horz_count/2].intensity +
				max_intensity)) / (2.0 * max_intensity));

	Resetfreq();
	generator_ptr->Seed();
}

void CA::Zeroseed()
{
	// Reset all the rows
	source_row = rowbuffer[0];
	sourcerowindex = 0;
	target_row = rowbuffer[1];
	targetrowindex = 1;
	past_row = rowbuffer[MEMORY - 1];
	pastrowindex = MEMORY - 1;
	wave_source_row = waverowbuffer[0];
	wavesourceindex = 0;
	wave_target_row = waverowbuffer[1];
	wavetargetindex = 1;
	for (int i = 0; i < GetSystemMetrics(SM_CXSCREEN); i++)
	{
		source_row[i] = 0;
		past_row[i] = 0;
		wave_source_row[i].intensity = 0.0;
		wave_source_row[i].velocity = 0.0;
		if (type_ca != CA_STANDARD && type_ca != CA_REVERSIBLE)
			source_row[i] = (int) ((MAX_COLOR *
				(wave_source_row[i].intensity +
				max_intensity)) / (2.0 * max_intensity));

	}
	Resetfreq();
	generator_ptr->Seed();
}

void CA::SyncRows()
{
/* This is to correct the fact that we have 16 rows and only
	row_buffer[0] gets seeded with all 640 cells. So if the size
	changes and we aren't at row_buffer[0], we need to copy what we
	currently have back to row_buffer[0]. Therefore row_buffer[0]
	acts as a kind of memory when we change the size around. */

	// First, see if we are at row_buffer[0] in the first place
	if (source_row != rowbuffer[0])
	{
		// If not, copy what we have to row_buffer[0]
		for (short i = 0; i < horz_count; i++)
		{
			rowbuffer[0][i] = source_row[i];
			rowbuffer[MEMORY-1][i] = past_row[i];
		}
		// Now reset everything to the top
		source_row = rowbuffer[0];
		sourcerowindex = 0;
		target_row = rowbuffer[1];
		targetrowindex = 1;
		past_row = rowbuffer[MEMORY-1];
		pastrowindex = MEMORY-1;
	}
	// See if we are at waverowbuffer[0]
	if (wave_source_row != waverowbuffer[0])
	{
		// If not, copy what we have to waverowbuffer[0]
		for (short i = 0; i < horz_count; i++)
		{
			waverowbuffer[0][i].intensity = wave_source_row[i].intensity;
			waverowbuffer[0][i].velocity = wave_source_row[i].velocity;
		}
		// Now reset everything to the top
		wave_source_row = waverowbuffer[0];
		wavesourceindex = 0;
		wave_target_row = waverowbuffer[1];
		wavetargetindex = 1;
	}
}

void CA::Locate(int itile_number, int dmaxx, int dmaxy, int CA_count_per_edge)
{
	int dwidth, dheight, cwidth, cheight; //display and cell window sizes
	int xpos,ypos;

/* move all rows back to default start position, but retain all the current
	information in them, so if we are increasing the size, we have the
	rest of the seeded rowbuffer[0] to look at initially */

	SyncRows();

	tile_number = itile_number;
	dwidth = dmaxx - 2;
	dheight = dmaxy - 2;

	cwidth = (dwidth  - (BORDER * (CA_count_per_edge - 1))) /
		CA_count_per_edge;
	cheight = (dheight - (BORDER * (CA_count_per_edge - 1))) /
		CA_count_per_edge;

	// Make it a factor of blt_lines
	cheight = (cheight / BLT_LINES) * BLT_LINES;

	xpos = tile_number % CA_count_per_edge;
	ypos = tile_number / CA_count_per_edge;

	minx = 1 + xpos*(cwidth + BORDER);
	miny = 1 + ypos*(cheight + BORDER);
	maxx = minx + cwidth - 1;
	maxy = miny + cheight - 1;

	horz_count = maxx - minx + 1;
	vert_count = maxy - miny + 1;

	switch (calist_ptr->viewmode)
	{
		case IDC_DOWN_VIEW:   row_number = miny; break;
		case IDC_SCROLL_VIEW: row_number = maxy - BLT_LINES + 1; break;
		default : row_number = miny; break;
	}

	//Since you're starting a new size, reset the Stripe avoider
	sourcerowindex = 0;
	targetrowindex = 1;
	pastrowindex = MEMORY - 1;
	/* we need to reset the counter here so that it will be in sync
		with the advancement of rownumber */
	computed_lines =0;
}

void CA::StandardUpdate(HDC hdc)
{
	int i,leftindex;
	unsigned short nabe = 0;

	if(wrapflag)
		for (i = horz_count-radius; i < horz_count; i++)
		{
			nabe |= source_row[i];
			nabe <<= statebits;
		}
	for (i = 0; i<radius; i++)
	{
		nabe |= source_row[i];
		nabe <<= statebits;
	}
	nabe |= source_row[radius];
	target_row[0] = lookup[nabe];
	if (entropyflag)
	{
		cellcount++;
		freqlookup[nabe]++;
	}
	for (i = 1; i < horz_count-radius; i++)
	{
		nabe <<= statebits;
		nabe &= mask;
		nabe |= source_row[i+radius];
		target_row[i] = lookup[nabe];
		if (entropyflag)
		{
			cellcount++;
			freqlookup[nabe]++;
		}
	}
	if (wrapflag)
	{
		leftindex = 0;
		for (i = horz_count-radius; i < horz_count; ++i)
		{
			nabe <<= statebits;
			nabe &= mask;
			nabe |= source_row[leftindex];
			target_row[i] = lookup[nabe];
			if (entropyflag)
			{
				cellcount++;
				freqlookup[nabe]++;
			}
			leftindex++;
		}
	}
	else //not wrapflag
		for (i = horz_count-radius; i < horz_count; ++i)
		{
			nabe <<= statebits;
			nabe &= mask;
			target_row[i] = lookup[nabe];
			if (entropyflag)
			{
				cellcount++;
				freqlookup[nabe]++;
			}
		} // end wrapflag if else.

	if (generatorflag)
		generator_ptr->Step();
/* We let the generator, if it's on, write to the target row.  This
way you see the generator value, and this value is used as source
after the rows are rolled just below.*/

	Show(hdc);
}

void CA::Show(HDC hdc)
{
	short i;
// Store the target row values into the bitmap.
	switch(calist_ptr->viewmode)
	{
		case IDC_DOWN_VIEW:
			for (i = 0; i < horz_count; ++i)
				WBM->WBMOnlyPutPixel(minx + i, row_number,
					 colortable[target_row[i]]);
			row_number++;
			if (row_number > maxy)
				row_number = miny;
		break;
		case IDC_SCROLL_VIEW:
			//Do the bump first, if needed, so the image looks good.
			if (row_number == maxy - BLT_LINES + 1)
				WBM->WBMBumpup(minx, miny, maxx, maxy, BLT_LINES);
			for (i = 0; i < horz_count; ++i)
				WBM->WBMOnlyPutPixel(minx + i,row_number,
					 colortable[target_row[i]]);
			row_number++;    //Starts at maxy - BLT_LINES + 1
			if (row_number > maxy)
				row_number = maxy - BLT_LINES + 1;
			break;
		case IDC_WIRE_VIEW:
			for (i = 0; i < horz_count; ++i)
				WBM->WBMOnlyPutPixel(minx + i,miny +
				(int)(vert_count/2), colortable[target_row[i]]);
			WBM->WBMWireBlt(hdc, minx, miny+(int)(vert_count / 2),
				maxx, 1);
			break;
		case IDC_GRAPH_VIEW:
		{
			WBM->ClearSection(minx, miny, maxx, maxy);
			for (i = 0; i < horz_count; ++i)
					WBM->WBMOnlyPutPixel(minx + i,
					maxy - (int)((vert_count-1) *
					((float)(target_row[i])/ (states-1))),
					RGB(255, 255, 255));
			break;
		}
	}
	//Roll forward the source and target buffers.
	if (++sourcerowindex >= MEMORY)
		sourcerowindex = 0;
	if (++targetrowindex >= MEMORY)
		targetrowindex = 0;
	source_row = rowbuffer[sourcerowindex];//same as target_row, actually.
	target_row = rowbuffer[targetrowindex];
	if (cellcount > (calist_ptr->breedcycle * horz_count))
	 /*could overflow one of the freqlookup entries in the next row */
		Entropy();
	if (sourcerowindex == MEMORY - 1)
		Avoidstripes();
}

void CA::ReversibleUpdate(HDC hdc)
{
	int i,leftindex;
	unsigned short nabe = 0;
	unsigned char statesmask;

	statesmask = states-1; //assume states is a power of two
	if(wrapflag)
		for (i = horz_count-radius; i < horz_count; i++)
		{
			nabe |= source_row[i];
			nabe <<= statebits;
		}
	for (i = 0; i<radius; i++)
	{
		nabe |= source_row[i];
		nabe <<= statebits;
	}
	nabe |= source_row[radius];
	target_row[0] = (lookup[nabe] + states - past_row[0]) & statesmask;
	/* This is for reversibility.  The idea is that Newstate =
		Computedstate - Paststate MOD states.  To stay away from
		negatives, we add states - Paststate.  To do a fast MOD,
		we and with statesmask (which is states-1).  Note there
		could be a problem here if states is over 128.*/
	if (entropyflag)
	{
		cellcount++;
		freqlookup[nabe]++;
	}
	for (i = 1; i < horz_count-radius; i++)
	{
		nabe <<= statebits;
		nabe &= mask;
		nabe |= source_row[i+radius];
		target_row[i] = (lookup[nabe] + states - past_row[i]) &
			statesmask;
		if (entropyflag)
		{
			cellcount++;
			freqlookup[nabe]++;
		}
	}
	if (wrapflag)
	{
		leftindex = 0;
		for (i = horz_count-radius; i < horz_count; ++i)
		{
			nabe <<= statebits;
			nabe &= mask;
			nabe |= source_row[leftindex];
			target_row[i] = (lookup[nabe] + states - past_row[i]) &
				statesmask;
			if (entropyflag)
			{
				cellcount++;
				freqlookup[nabe]++;
			}
			leftindex++;
		}
	}
	else
		for (i = horz_count-radius; i < horz_count; ++i)
		{
			nabe <<= statebits;
			nabe &= mask;
			target_row[i] = (lookup[nabe] + states - past_row[i]) &
				statesmask;
			if (entropyflag)
			{
				cellcount++;
				freqlookup[nabe]++;
			}
		}

	if (++pastrowindex >= MEMORY) //for reversible_flag
		pastrowindex = 0;
	past_row = rowbuffer[pastrowindex];

	if (generatorflag)
		generator_ptr->Step();
/* We let the generator, if it's on, write to the target row.  This
way you see the generator value, and this value is used as source
after the rows are rolled just below.*/

	Show(hdc);
}

void CA::WaveUpdate(HDC hdc)
{
	for (short i = 1; i < horz_count-1; i++)
		(this->*UpdateCell)(i-1, i, i+1);
// Do the same thing at the ends if wrapflag, else set to 0.0.
	if(wrapflag)
	{
		// left end update
		(this->*UpdateCell)(horz_count-1, 0, 1);
	// right end update
		(this->*UpdateCell)(horz_count-2, horz_count-1, 0);
	}
	else
	{
	// left end, wrap off
		wave_target_row[0].intensity = 0.0;
		wave_target_row[0].velocity = 0.0;
	// right end, wrap off
		wave_target_row[horz_count-1].intensity = 0.0;
		wave_target_row[horz_count-1].velocity = 0.0;
	}
	WaveUpdateStep(hdc);
}

void CA::WaveUpdate2(HDC hdc)
{
	for (short i = 2; i < horz_count-2; i++)
		(this->*UpdateCell2)(i-2, i-1, i, i+1, i+2);
// Do the same thing at the ends if wrapflag, else set to 0.0.
	if(wrapflag)
	{
	// left end update
		(this->*UpdateCell2)(horz_count-2, horz_count-1, 0, 1, 2);
		(this->*UpdateCell2)(horz_count-1, 0, 1, 2, 3);
	// right end update
		(this->*UpdateCell2)(horz_count-3, horz_count-2,
			horz_count-1, 0, 1);
		(this->*UpdateCell2)(horz_count-4, horz_count-3,
				horz_count-2, horz_count-1, 0);
	}
	else
	{
	// left end, wrap off
		wave_target_row[0].intensity = 0.0;
		wave_target_row[0].velocity = 0.0;
		wave_target_row[1].intensity = 0.0;
		wave_target_row[1].velocity = 0.0;
	// right end, wrap off
		wave_target_row[horz_count-1].intensity = 0.0;
		wave_target_row[horz_count-1].velocity = 0.0;
		wave_target_row[horz_count-2].intensity = 0.0;
		wave_target_row[horz_count-2].velocity = 0.0;
			//Pass RR same as R: zero.
	}
	WaveUpdateStep(hdc);
}

void CA::WaveUpdateStep(HDC hdc)
{
	if (generatorflag)
		generator_ptr->Step();
/* We let the generator, if it's on, write to the wave target row.  This
way you see the generator value, and this value is used as source
after the rows are swapped just below.*/

	/* Now clamp all intensities and velocities so they won't get too big
		and write all the rows to the bitmap */

	for (short i = 0; i < horz_count; i++)
	{
/* You might think of setting the velocity to 0.0 when the
intensity gets clipped, but this seems not be a good idea, as then
the synchronization doesn't develop and you don't get nice waves.*/
//		RealWrap(wave_target_row[i].intensity, -max_intensity, max_intensity);
//		RealWrap(wave_target_row[i].velocity,-max_velocity, max_velocity);
/* map the (intensity to max_intensity) range to (0 to 256) color range
		and store the result in target_row */
		if (!(calist_ptr->showvelocity))
			target_row[i] = (int) ((MAX_COLOR *
				(wave_target_row[i].intensity +
				max_intensity)) / (2.0 * max_intensity));
		else
			target_row[i] = (int) ((MAX_COLOR *
				(wave_target_row[i].velocity +
				max_velocity)) / (2.0 * max_velocity));

		if (entropyflag)
		{
			cellcount++;
			freqlookup[target_row[i]]++;
		}
	}
	switch (showmode)
	{
		case BOTH_SHOW:
			break;
		case ODD_SHOW:
			for (i = 0; i < horz_count-1; i += 2)
				target_row[i] = target_row[i+1];
			if (horz_count & 1) //last index is even
				target_row[horz_count-1] = target_row[horz_count-2];
			break;
		case EVEN_SHOW:
			for (i = 1; i < horz_count; i += 2)
				target_row[i] = target_row[i-1];
			break;
	}

	if (++wavesourceindex >= 2)
		wavesourceindex = 0;
	if (++wavetargetindex >= 2)
		wavetargetindex = 0;
	wave_source_row = waverowbuffer[wavesourceindex];
		//same as wave_target_row, actually
	wave_target_row = waverowbuffer[wavetargetindex];

//Now step the time.
	time += time_step;
	if (time > 2.0*PI)
		time -= 2.0*PI;

	Show(hdc);
}

void CA::WaveVelInt1(int l, int c, int r)
{
/*  The wave equation is utt = 1(/speed^2) * uxx, where
"speed" is the maximum speed of the wave in the medium.  In terms of
the simulation, I use a space_step and a time_step: one space_step per
cell to the left or right, one time_step per update from one generation
to the next.  Speed thus is space_step/time_step.  It simplifies the
compuation to assume speed is 1.0, and to scale space_step and time_step
to make this true.  It turns out, I only get stability if space_step is
time_step * 2.0.
	If speed is 1.0, the wave equation is just
utt = uxx.
We implement this by keeping
track of u, also known as intensity, and of ut, also known as v or velocity.
	The update rule is thus:
new v = v + dt * uxx
	We can approximate uxx as (L - 2C +R)/2.0, so get
new v = v + (dt/dx*dx) (L - 2C + R).
	Pre-compute (dt/dx*dx), also known as (time_step/space_step*space_step),
and store it as accelration_multiplier.
	In OLDWAVE1, simply update
new u = u + dt * new v.
	Note by the way, that it is important to use new v in the second
equation, otherwise you get instability, with the odd and even cells
forever 180 degrees out of phase.  In other words, new u = u + dt * v
gives instability.
	Another issue is that we need to keep the velocities and intensities
within a reasonable range, otherwise they can drift up into invalid
floating-point values for certain initial conditions.  The choice for
this simulation is to "wrap" the values, meaning that you have a max
and a min value and if a value goes above max you wrap it down to the
same amount above the min, and if a value goes below the min you wrap
it around to the same amount below the max.  Clamping at the max or
min value works too, but is less lively to watch.
	It turns out even this method does not work well, we need to go
to a higher-order method, a second-order way.  The
idea is to expand u(t+dt) as a Taylor series in dt, and get
 new u = u + dt ut + (dt*dt/2)utt.
 Replace ut by v, and replact utt by uxx and get
 new u = u + dt v +  (dt*dt/2)uxx.
 When you compute (dt/dx*dx)(L - 2C + R) for the v update, save this as
 dtutt, so now have
 new u = u + dt v + (dt/2) dtutt;
*/
//#define OLDWAVE1
#ifdef OLDWAVE1
	wave_target_row[c].velocity = wave_source_row[c].velocity +
		acceleration_multiplier *(
		wave_source_row[l].intensity -
		2.0*wave_source_row[c].intensity+
		wave_source_row[r].intensity);
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
//	wave_target_row[c].intensity = wave_source_row[c].intensity +
//		time_step * wave_source_row[c].velocity;  // Old, unstable way.
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity; //Stable way.
	RealWrap(wave_target_row[c].intensity, -max_intensity, max_intensity);
#else  //Second order way.
	Real dtutt = 	acceleration_multiplier *(
		wave_source_row[l].intensity -
		2.0*wave_source_row[c].intensity+
		wave_source_row[r].intensity);
	wave_target_row[c].velocity = wave_source_row[c].velocity + dtutt;
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
// Old, unstable way:
//	wave_target_row[c].intensity = wave_source_row[c].intensity +
//		time_step * wave_source_row[c].velocity + half_time_step*dtutt;
// Stable way:
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity + half_time_step*dtutt;
	RealWrap(wave_target_row[c].intensity, -max_intensity, max_intensity);
#endif //OLDWAVE1
}

void CA::UlamWave(int l, int c, int r)
{
	Real cldiff = wave_source_row[c].intensity -
		wave_source_row[l].intensity;
	Real rcdiff =  wave_source_row[r].intensity -
		wave_source_row[c].intensity;
	Real dtutt = 	acceleration_multiplier *(
		wave_source_row[l].intensity -
		2.0*wave_source_row[c].intensity+
		wave_source_row[r].intensity +
		spring_multiplier * ( rcdiff*rcdiff - cldiff*cldiff )	);
	wave_target_row[c].velocity = wave_source_row[c].velocity + dtutt;
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity + half_time_step*dtutt;
	CLAMP(wave_target_row[c].intensity, -max_intensity, max_intensity);
}

void CA::WaveVelInt2(int ll, int l, int c, int r, int rr)
{
  /*  Here we are going to do the same thing as WaveVelInt1, except
  we look at five neighbors.  The second differences formula for
  uxx = (-LL + 16 L -30 C + 16 R - RR)/(12 dx*dx).
*/
//#define OLDWAVE2
#ifdef OLDWAVE2
	wave_target_row[c].velocity = wave_source_row[c].velocity +
		 acceleration_multiplier *(
		- wave_source_row[ll].intensity +
		16.0*wave_source_row[l].intensity -
		30.0*wave_source_row[c].intensity+
		16.0*wave_source_row[r].intensity -
		wave_source_row[rr].intensity);
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity;
	RealWrap(wave_target_row[c].intensity, -max_intensity, max_intensity);
#else
	Real dtutt = 	acceleration_multiplier *(
		- wave_source_row[ll].intensity +
		16.0*wave_source_row[l].intensity -
		30.0*wave_source_row[c].intensity+
		16.0*wave_source_row[r].intensity -
		wave_source_row[rr].intensity);

	wave_target_row[c].velocity = wave_source_row[c].velocity + dtutt;
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity + half_time_step*dtutt;
	RealWrap(wave_target_row[c].intensity, -max_intensity, max_intensity);
#endif //OLDWAVE2
}

void CA::WaterWave(int ll, int l, int c, int r, int rr)
{
/* Russell Herman, "Solitary Waves," THE AMERICAN SCIENTIST Vol 80, July-Aug
1992, pp. 350-361: says use du/dt = -6 u * du/dx - d^3u/dx^3.  Let's try
this today and also try taking space_step into account.  In this case the
derivatives are
 ux: (R - L)/(2 * deltax);
 uxx: (R - 2C + L)/ deltax^2
 uxxx: (RR -2R + 2L - LL) / (2.0 * deltax^3)
 */

	wave_target_row[c].velocity =
		wave_source_row[c].velocity +
		(time_step ) *
		(
			-6.0 * wave_source_row[c].intensity *
			(
				(wave_source_row[r].intensity -
				wave_source_row[l].intensity) / (2.0 * space_step)
			) -
			(
				(wave_source_row[rr].intensity -
				2.0 * wave_source_row[r].intensity +
				2.0 * wave_source_row[l].intensity -
				wave_source_row[ll].intensity) /
					(2.0 * pow(space_step, 3.0) )
			)
		);
 /*	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_source_row[c].velocity; */
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity;
	CLAMP(wave_target_row[c].intensity, -max_intensity, max_intensity);

}

void CA::Oscillator(int l, int c, int r)
{
/* The forced oscillator idea is from Feynman, Lectures on Physics,
Vol. 1, Chaps 23 & 24*
Inductance * d charge^2/dt^2 + Resistance * d charge/dt + charge/Capacitance
= Voltage(t).  What I call intensity, then, is charge, and what I call
velocity is current.
I * d velocity/dt + R * velocity + intensity/C = Driver(t).
d velocity /dt = (1/I)(-R*velocity - (1/C)intensity + Driver(t)), see
p. 23-6, or
current += (dt/Inductance)*
	(-Resistance*current - (1/Capacitance)*charge + Voltage(t));
We could also write, p. 23-7
dI/dt = (1/Inductance)*(-RI - (1/C) Integral of I to t +V(t).
where the integral of I to t is just the charge.
*/
//#define OLDWAYO
#ifndef OLDWAYO
	wave_target_row[c].velocity = wave_source_row[c].velocity +
		(time_step / mass)* (
		- friction_multiplier * wave_source_row[c].velocity -
		spring_multiplier * wave_source_row[c].intensity +
		driver_multiplier * cos(frequency_factor * time));
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity;
	CLAMP(wave_target_row[c].intensity, -max_intensity, max_intensity);
#else
		Real dtutt =
		(time_step / mass)* (
		- friction_multiplier * wave_source_row[c].velocity -
		spring_multiplier * wave_source_row[c].intensity +
		driver_multiplier * cos(frequency_factor * time));
	wave_target_row[c].velocity = wave_source_row[c].velocity + dtutt;
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity + half_time_step*dtutt;
	CLAMP(wave_target_row[c].intensity, -max_intensity, max_intensity);
#endif OLDWAYO
}

void CA::OscillatorWave(int l, int c, int r)
{
//#define OLDWAYOW
#ifdef OLDWAYOW
	wave_target_row[c].velocity = wave_source_row[c].velocity +
		(1.0/mass)*(time_step * (
		- friction_multiplier * wave_source_row[c].velocity -
		spring_multiplier * wave_source_row[c].intensity +
		driver_multiplier * cos(frequency_factor * time) )+
		acceleration_multiplier * (
		wave_source_row[l].intensity -
		2.0*wave_source_row[c].intensity +
		wave_source_row[r].intensity));
/*	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_source_row[c].velocity;  */
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity;
	CLAMP(wave_target_row[c].intensity, -max_intensity, max_intensity);
#else
	Real dtutt = (1.0/mass)*( time_step * (
		- friction_multiplier * wave_source_row[c].velocity -
		spring_multiplier * wave_source_row[c].intensity +
		driver_multiplier * cos(frequency_factor * time)) +
		acceleration_multiplier * (
		wave_source_row[l].intensity -
		2.0*wave_source_row[c].intensity +
		wave_source_row[r].intensity));
	wave_target_row[c].velocity = wave_source_row[c].velocity +
		dtutt;
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
//	wave_target_row[c].intensity = wave_source_row[c].intensity +
//		time_step * wave_target_row[c].velocity;
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity + half_time_step*dtutt;
	CLAMP(wave_target_row[c].intensity, -max_intensity, max_intensity);
#endif //OLDWAYOW
}

void CA::DiverseOscillator(int l, int c, int r)
{
	wave_target_row[c].velocity = wave_source_row[c].velocity +
		(time_step/mass) * (
		- friction_multiplier * wave_target_row[c].friction_multiplier *
     	wave_source_row[c].velocity -
		spring_multiplier*wave_target_row[c].spring_multiplier *
     	wave_source_row[c].intensity +
		driver_multiplier*wave_target_row[c].driver_multiplier *
		 cos(frequency_factor * wave_target_row[c].frequency_multiplier* time));
/*	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_source_row[c].velocity; */
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity;
	CLAMP(wave_target_row[c].intensity, -max_intensity, max_intensity);
}

void CA::DiverseOscillatorWave(int l, int c, int r)
{
//#define OLDWAYDOW
#ifdef OLDWAYDOW
	wave_target_row[c].velocity = wave_source_row[c].velocity +  (1/mass)*(
   	time_step* (
		- friction_multiplier * wave_target_row[c].friction_multiplier *
     	wave_source_row[c].velocity -
		spring_multiplier*wave_target_row[c].spring_multiplier *
     	wave_source_row[c].intensity +
		driver_multiplier*wave_target_row[c].driver_multiplier *
		 cos(frequency_factor * wave_target_row[c].frequency_multiplier* time)
       ) +
		acceleration_multiplier *(
		wave_source_row[l].intensity -
		2.0*wave_source_row[c].intensity+
		wave_source_row[r].intensity));
 /*	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_source_row[c].velocity; */
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity;
	CLAMP(wave_target_row[c].intensity, -max_intensity, max_intensity);
#else
	Real dtutt = (1.0/mass)*(
  	time_step* (
		- friction_multiplier * wave_target_row[c].friction_multiplier *
     	wave_source_row[c].velocity -
		spring_multiplier*wave_target_row[c].spring_multiplier *
     	wave_source_row[c].intensity +
		driver_multiplier*wave_target_row[c].driver_multiplier *
		 cos(frequency_factor * wave_target_row[c].frequency_multiplier* time)
       ) +
		acceleration_multiplier *(
		wave_source_row[l].intensity -
		2.0*wave_source_row[c].intensity+
		wave_source_row[r].intensity));
	wave_target_row[c].velocity = wave_source_row[c].velocity +  dtutt;
	CLAMP(wave_target_row[c].velocity,-max_velocity, max_velocity);
	wave_target_row[c].intensity = wave_source_row[c].intensity +
		time_step * wave_target_row[c].velocity + half_time_step*dtutt;
	CLAMP(wave_target_row[c].intensity, -max_intensity, max_intensity);
#endif //OLDWAYDOW
}


void CA::HeatInt1(int l, int c, int r)
{
/* In the past, I've believed the heat equation says uxx=0, but it's
more accurate to say it says ut = uxx.
	First, crude, idea to try and make uxx = 0 .  With one neighbor, this
means (-2C + L + R) = 0, or C = (L+R)/2. This gives gross instability.
C = L+C+R / 3 works better.
	If we bring in the ut=uxx formulation, we have something like
C += (dt/dx*dx)(L -2C + R) as a first attempt.  This makes terrible
results, see Masatake Mori, THE FINITE ELMENT METHOD AND ITS APPLICATIONS,
PP.105, 106 for explanation. (If all are zero and C is plus, then C
jumps to negative!)
	We can get a different formulation, writing lam for (dt/dx*dx).
(nC - C)/dt = (nL -2 nC +nR)/dxdx
nC = C + lam(nL - 2nC + nR)
(1 + 2*lam)nC = lam nL + C + lam nR
replace nL by L and nR by R (!!!!) and get
nC = (lam L + C + lam R)/(1 + 2*lam).
This is a nice formula because it incorporates dt and dx, and it is like
a weighted averaging scheme.
	In use, it is more stable if you pick a large space_step, the idea
being that lam, A.K.A. acceleration_parameter, shouldn't be very large.
	To drive the heating, I add dt+heat_inc to u each step as well.
	Rather than clamping, we wrap.
*/
//#define SIMPLE
#ifdef SIMPLE
	wave_target_row[c].intensity = heat_inc +
		(wave_source_row[l].intensity +
		wave_source_row[c].intensity +
		wave_source_row[r].intensity) / 3.0;
	if (wave_target_row[c].intensity > max_intensity)
		wave_target_row[c].intensity -= 2.0*max_intensity;//= -max_intensity;
	//Assume heat_inc is positive, so only need to wrap at top.
#else
	wave_target_row[c].intensity = (
		acceleration_multiplier * wave_source_row[l].intensity +
		wave_source_row[c].intensity +
		acceleration_multiplier * wave_source_row[r].intensity) /
			(1 + 2.0 * acceleration_multiplier) +
		time_step * heat_inc;
	wave_target_row[c].velocity = (wave_target_row[c].intensity -
		wave_source_row[c].intensity)/time_step; //Calcualte just for graphing.
	RealWrap(wave_target_row[c].intensity, -max_intensity, max_intensity);
#endif
}

void CA::HeatInt2(int ll, int l, int c, int r, int rr)
{
/*  We can try a crude average for SIMPLE.
	For the complex case, use the five neighbor second difference,
uxx =	(-LL + 16 L -30 C + 16 R - RR)/(12 dx*dx)
	We can get a different formulation, writing lam for (dt/12*dx*dx).
(nC - C)/dt = (-nLL + 16 nL -30 nC + 16 nR - nRR)/12*dxdx
nC = C + lam(-nLL + 16 nL -30 nC + 16 nR - nRR)
(1 + 30*lam)nC = C + lam (-nLL + 16nL  +  16nR - nRR )
replace nL by L and nR by R  nLL by LL and nRR by RR(!!!!) and get
nC = (C + lam(-LL + 16 L + 16 R - RR))/(1 + 30*lam).
	This doesn't seem to be very stable.
*/
#ifdef SIMPLE
	wave_target_row[c].intensity = heat_inc +
		(wave_source_row[ll].intensity +
		wave_source_row[l].intensity +
		wave_source_row[c].intensity +
		wave_source_row[r].intensity +
		wave_source_row[rr].intensity) / 5.0;
	if (wave_target_row[c].intensity > max_intensity)
		wave_target_row[c].intensity -= 2.0*max_intensity;
	//Assume heat_inc is positive so only wrap at top.  You
	//may possibly get an intensity < -max_intensity, but this
	//will be clamped back to -max_intensity in RealWrap in WaveUpdate.
#else
	wave_target_row[c].intensity = (wave_source_row[c].intensity +
		acceleration_multiplier * (
		-wave_source_row[ll].intensity +
		 16.0*wave_source_row[l].intensity  +
		 16.0*wave_source_row[r].intensity -
		 wave_source_row[rr].intensity) ) /
			(1 + 30.0 * acceleration_multiplier) +
		time_step * heat_inc;
	wave_target_row[c].velocity = (wave_target_row[c].intensity -
		wave_source_row[c].intensity)/time_step; //Calcualte just for graphing.
	if (wave_target_row[c].intensity > max_intensity)
		wave_target_row[c].intensity -= 2.0*max_intensity;
	if (wave_target_row[c].intensity < -max_intensity)  //In case if neg inc
		wave_target_row[c].intensity += 2.0*max_intensity;
#endif
}

//---------------------
void CA::Changeradiusandstates(int iradius, int istates)
{
	/* Waves always have 256 states with radius of 1.
	We also give them 256 nabeoptions, and just view the state as
	the nabecode.  (Be careful never to try and allocate a Lookuptable for
	256 states radius 1.)*/
	if (type_ca != CA_STANDARD && type_ca != CA_REVERSIBLE)
	{
		if (states != 256 && //You are changing out of STANDARD or REVERSIBLE
			states != 0) //And this is not your first call to this method
		{
			oldradius = radius;
			oldstates = states;
		}
		radius = 1;
		states = 256;
		nabeoptions = 256; //Do this so I can do Entropy on Waves.
		nabesize = 0;
		return;
	}
	//At this point you must be STANDARD or REVERSIBLE.
	//If you are changing out of a WAVE type, put good args in.
	if (states == 256)
	{
		iradius = oldradius;
		istates = oldstates;
	}
	//If an arg is -1, this means you're not changing this one.
	if (iradius == -1)
		iradius = radius;
	if (istates == -1)
		istates = states;
/* if you just want to change just radius or states, put a -1 in the
	field you don't want to change */
	if (radius == iradius && states == istates)
		return;

	if ( (istates > 16) || (iradius==2 && istates>4) ||
		 (iradius==3 && istates>4) ||
		(iradius>=4 && istates > 2) || (iradius > 6))
		return;  //No can do!

	radius = iradius;
	states = istates;
	statebits = log(states)/log(2.0); //log base two is number of bits.
	nabesize = 1 + 2*radius;
	nabeoptions = pow(states, nabesize);
	mask = nabeoptions - 1;
	Resetfreq();
	Lambdalookup(); //Randomize the lookup table
	Setcolortable();
	Seed();
	score = 0.0;	// Reset the score
}

/*  I use all combos where lookup is less than or equal to 16K

States	2		4			8			16			32
Radius
1			8		64			512		4 K		32 K
2			32		1 K		32 K		1 Meg		32 Meg
3			128		16 K		2 Meg		256 Meg	32 Gig
4			512	256 K		16 Meg	4 Gig		32 Ter
5			2 K	4 Meg		8 Gig		16 Ter	32 Pet
6			8 K	64 Meg	512 Gig	4 Pet		32 Ex
7			32 K	1 Gig		32 Ter	1 Ex		32 K Ex


*/

/* This checks if the CAs are compatible with eachother.
	Since Standard and Reversible both use lookuptable, they can
	be exchange lookuptable information.
	Waves and Heatwaves can exchange velocity and intensity information */
unsigned char Compatible(int type1, int type2)
{
	if ((type1 == CA_STANDARD || type1 == CA_REVERSIBLE) &&
		(type2 == CA_STANDARD || type2 == CA_REVERSIBLE))
		return 1;
	if ((type1 != CA_STANDARD && type1 != CA_REVERSIBLE) &&
		(type2 != CA_STANDARD && type2 != CA_REVERSIBLE))
		return 1;
	return 0;
}

/* This copies everything! I replaced Copytable with this version.*/
void CA::CopyCA(CA *goodCA)
{
	unsigned short i;

	if (!Compatible(type_ca,goodCA->type_ca))
		return;

	Changeradiusandstates(goodCA->radius, goodCA->states);
	for(i=0; i<nabeoptions; i++)
		lookup[i] = goodCA->lookup[i];

	for (i = 0; i < GetSystemMetrics(SM_CXSCREEN); i++)
	{
		source_row[i] = goodCA->source_row[i];
		past_row[i] = goodCA->past_row[i];
		wave_source_row[i].intensity = goodCA->wave_source_row[i].intensity;
		wave_source_row[i].velocity = goodCA->wave_source_row[i].velocity;
	}

	lambda = goodCA->lambda;
	entropy= goodCA->entropy;
	target_entropy = goodCA->target_entropy;

	time_step = goodCA->time_step;
	space_step = goodCA->space_step;
	band_count = goodCA->band_count;
	max_intensity = goodCA->max_intensity;
	max_velocity = goodCA->max_velocity;
	friction_multiplier = goodCA->friction_multiplier;
	spring_multiplier = goodCA->spring_multiplier;
	mass = goodCA->mass;
	driver_multiplier = goodCA->driver_multiplier;
	frequency_multiplier = goodCA->frequency_multiplier;
	Adjust_acceleration_multiplier();

	for (i = 0; i <= MAX_COLOR; i++)
		colortable[i]= goodCA->colortable[i];
}

COLORREF CA::Color_look(int numb)
{
	return colortable[lookup[numb]];
}

void CA::Setcolortable()
{
	COLORREF ca_wave_color;
	int bandwidth;
	int i;
	// Use states as The number of color bands I want to have.
	HDC hdc;

	hdc = GetDC(hwnd);
	if (Gettype() == CA_STANDARD || Gettype() == CA_REVERSIBLE)
	{
		colortable[0] = GetNearestColor(hdc, RGB(0,0,0));
		colortable[1] = GetNearestColor(hdc, RGB(255,255,255));
		for (i = 2; i <= MAX_COLOR; i++)
			colortable[i]= GetNearestColor(hdc, RGB(Random(255),Random(255),Random(255)));
	}
	else //CA_TYPE_WAVE case
	{
 /*The idea here is to set the medium values to black,
	and to band the other colors in between */

		bandwidth = MAX_COLOR / band_count; //The width of a band
		for (i = 0; i <= MAX_COLOR; i++)
		{
//#define GRAYSCALE  //Keep this in, November 9, 1994.  This works good
// with 256 color mode
#ifndef GRAYSCALE
			if ((i/bandwidth)*bandwidth == i)
// if i is a multiple of bandwidth pick a color for the next band.
				ca_wave_color = GetNearestColor(hdc, RGB(Random(255),
						Random(255),Random(255)));
			//Make the low values blue.
			 if (i < bandwidth)
					colortable[i] = GetNearestColor(hdc, RGB(0,0,255));
			//Make the high values red.
			else if (i > (MAX_COLOR - bandwidth - 1))
				colortable[i] = GetNearestColor(hdc, RGB(255,0,0));
			 //Make the medium values all black
			else if ((i > ((MAX_COLOR/2) - bandwidth)) &&
				(i < ((MAX_COLOR/2) + bandwidth)))
				colortable[i] = GetNearestColor(hdc, RGB(0,0,0));
			else // Use the band color
				colortable[i]= ca_wave_color;
#else  //GRAYSCALE
			colortable[i] = GetNearestColor(hdc, RGB(i,i,i));
#endif //GRAYSCALE
		}
	}
	ReleaseDC(hwnd, hdc);
}

void CA::Setoldcolor()
{

	colortable[0] = RGB(0,0,0);
	if (states == 2)
		colortable[1] = RGB(255,255,255);
	else
	{
		colortable[1] = RGB(0,0,255);
		colortable[2] = RGB(0,255,0);
		colortable[3] = RGB(0,255,255);
		colortable[4] = RGB(255,0,0);
		colortable[5] = RGB(255,0,255);
		colortable[6] = RGB(255,255,0);
		colortable[7] = RGB(100,100,100);
		colortable[8] = RGB(40,40,40);
		colortable[9] = RGB(100,100,255);
		colortable[10] = RGB(100,255,100);
		colortable[11] = RGB(100,255,255);
		colortable[12] = RGB(255,100,100);
		colortable[13] = RGB(255,100,255);
		colortable[14] = RGB(255,255,100);
		colortable[15] = RGB(255,255,255);
	}
}

void CA::Changelambda(float increment)
{
	unsigned short i;
	float oldlambda, flip_prob;

	oldlambda = lambda;
	lambda += increment;
	if (lambda > MAX_LAMBDA)
		lambda = MAX_LAMBDA;
	if (lambda < MIN_LAMBDA)
		lambda = MIN_LAMBDA;

	if (increment > 0) //increase lambda
	{
		flip_prob = (lambda - oldlambda)/(1.0 - oldlambda);
/* Because lambda = ((1.0 - oldlambda)*flip_prob*nabeoptions+
oldlambda*nabeoptions) / nabeoptions; or
lambda = (1-oldlambda)*flip_prob + oldlambda.*/
		for (i=0; i<nabeoptions; i++)
			if (!lookup[i] && Randomreal()< flip_prob )
				lookup[i] = 1 + Random(states - 1);
	}
	else //decrease lambda
	{
		flip_prob = (oldlambda - lambda)/(oldlambda);
/* Because lambda = (oldlambda*nabeoptions -
oldlambda*nabeoptions*flip_prob) / nabeoptions; or
lambda = oldlambda - oldlamda*flip_prob.*/
		for (i=0; i<nabeoptions; i++)
			if (lookup[i] && Randomreal()< flip_prob )
				lookup[i] = 0;
	}
	Computeactual_lambda();
	Resetfreq();
}

void CA::Changeentropy(double newentropy)
{
	target_entropy = newentropy;
}

void CA::Become_child_of(CA *mom, CA *dad)
{
	unsigned short i, breakpoint, momfirst;

	if (!Compatible(mom->type_ca, dad->type_ca))
		return;

	Settype(mom->type_ca);

	if (mom->type_ca == CA_STANDARD ||
		 mom->type_ca == CA_REVERSIBLE)
	{
		Changeradiusandstates(mom->radius, mom->states);
		if (dad->nabeoptions < nabeoptions) //Engulf dad.
		{
			for(i = 0; i<dad->nabeoptions; i++)
				lookup[i] = dad->lookup[i] % states;
			for (i = dad->nabeoptions; i<nabeoptions; i++)
				lookup[i] = mom->lookup[i];
		}
		else //crossover with dad
			{
				momfirst = Random(2);
				breakpoint =(unsigned short)Longrandom(nabeoptions);
				if (momfirst)
				{
					for (i=0; i<breakpoint; i++)
						lookup[i] = mom->lookup[i];
					for (i=breakpoint; i<nabeoptions; i++)
						lookup[i] = dad->lookup[i] % (unsigned char)states;
				}
				else
				{
					for (i=0; i<breakpoint; i++)
						lookup[i] = dad->lookup[i] % (unsigned char)states;
					for (i=breakpoint; i<nabeoptions; i++)
						lookup[i] = mom->lookup[i];
				}
			}
		for (i=0; i < states; i++)
		{
			if (Random(2))
				colortable[i] = mom->colortable[i];
			else
				colortable[i] = dad->colortable[i];
		}
		Computeactual_lambda();
	}
	else
	{
//---CA_wave
		time_step = (mom->time_step +
			dad->time_step) / 2.0;
		space_step = (mom->space_step +
			dad->space_step) / 2.0;
		Adjust_acceleration_multiplier();
	}
}

void CA::Entropy()
{
	unsigned short i;
	double sum = 0.0;
	double temp;
#define FAST_ENTROPY 1
#ifdef FAST_ENTROPY
	double logcellcount;
#else
	double probability;
#endif //FAST_ENTROPY

	if (entropyflag)
	{
// This is based on the definition of entropy as the negtative sum over
// the cases of the probability of each case times the log base two
// of that probability.  Use negative because the log of a probability,
// which is less than 1, is less than 1.  See next comment for why
// we don't have to divide each term by LOG2.
		if (!cellcount) //just in case
		{
			entropy = 0.0;
			return;
		}
#ifdef FAST_ENTROPY
/*This is an idea for a faster computation based on some twiddling
with the formula.  Instead of 1/log(nabeoptions) * Sum 1 to nabeoptions of
- freqlookup[i]/cellcount* log(freqlookup[i]/cellcount); I have
1/cellcount*log(nabeoptions) * Sum 1 to nabeoptions of - freqlookup[i] * (
log(freqlookup[i]) - log(cellcount)); and I store log(cellcount) at the
start to cut down on the computations.*/
		logcellcount = log((double)cellcount); //We know cellcount != 0.
		for (i=0; i<nabeoptions; i++)
		{
			// Stay away from log(0.0).
			if (freqlookup[i])
				sum -= (double)freqlookup[i] * ( log(freqlookup[i]) -
					logcellcount );
		}

		if (nabeoptions <= 1) //just in case
			entropy = 0.0;
		else
			entropy = sum / ((double)cellcount * log((double)nabeoptions) );
#else //The tried and true method
		for (i=0; i<nabeoptions; i++)
		{
			probability = (double)(freqlookup[i]) / (double)cellcount;
			// Stay away from log(0.0).
			if (probability != 0.0)
				sum -= probability * (log(probability));
			freqlookup[i] = 0; //reset for the next time.
		}
		// Note that I give the colony's target entropy as a percent between
		// 0 and 1.  The max actual entropy per bug action is
		// log-base-two of the number of possible outcomes, so the
		// percent entropy is actual entropy value of the sum
		// over the max possible entropy per bug action.  The
		// divisions by LOG2 (to scale to base two logarithm) cancel
		// each other out.
		if (nabeoptions <= 1) //just in case
			entropy = 0.0;
		else
			entropy = sum / log((double)nabeoptions);
#endif //FAST_ENTROPY

//		fitness = 1.0 - fabs(entropy - target_entropy);
//		calist_ptr->Sortfitness(); /*update CAlist's bestinlist and
//			secondbestinlist indices*/
//		if (fitness < HIGH_FITNESS)
//			calist_ptr->Evolve(*this);

	} //End of entropyflag case.
	Resetfreq();

	if (calist_ptr->breedflag)	// Only keep score if breed cycle is enabled
	{
		temp = target_entropy - entropy;
		if (temp < 0)
			temp *= -1.0;	/* I would use abs(target_entropy-entropy), but abs()
								uses integers only */
		score += entropy_bonus * (1.0 - (temp));	/* the closer you are to
			the target_entropy, the more bonus you get */
	}

	if (calist_ptr->FocusCA() == this)
		update_flag = 1; //need to update dialog boxes
}

void CA::Avoidstripes()
{
	unsigned char stripe = 1;
	int i, j, k, lightcone, rowi;

	//First check if you even care if you fail the stripe test or not.
	if (!( (calist_ptr->breedflag && fail_stripe) ||
		calist_ptr->stripekillflag ||
		calist_ptr->stripeseedflag ) )
			return;

	//First check if the row is all the same color
	for (i=radius; i<horz_count-radius; i++)
		if (source_row[i] != source_row[radius])
		{
			stripe = 0; //Not all the same in row
			break; //Bail from this i loop.
		}
	if (stripe) //Change the CA if it fails the test
	{
		if (type_ca != CA_REVERSIBLE && calist_ptr->stripeseedflag)
			Seed();

		if (calist_ptr->stripekillflag)
			Lambdalookup();
		else
		{
			if (calist_ptr->breedflag)	// Only keep score if breed cycle is enabled
				score += fail_stripe;	// penalty for fail stripe
		}
		if (calist_ptr->FocusCA() == this)
			update_flag = 1;					// update combo boxes
		return;
	}
	/* Now check for each of the rows in the rowbuffer if
		that row is a displacement of the sourcerow.  The possible
		displacements range up to radius * steps between rows. */
	for (j = sourcerowindex - 1; j>=0; j--)
	{ //Note that sourcerowindex = MEMORY - 1 when you call Avoidstripes.
		lightcone = radius * (sourcerowindex - j);
		for (k = -lightcone; k<= lightcone; k++)
		{
			stripe = 1; //Assume that it is a copy
	/* Now we want to compare all the cells in the row, with
	the rowbuffer row displaced by k cells.  Use rowi to
	index the displaced cells. The comments below trace through
	what would happen in the special case UPOS k=1 and UNEG k=-1
	just to make sure*/
			if (k<0)
			{
				rowi = horz_count+k; //UNEG rowi=horz_count-1
				if (rowi<0)
					continue; /* For a very small window, it might happen
				that horz_count is smaller than lightcone, which is
				MEMORY-1 * radius at the most.  Allowing a negative rowi
				would cause a crash when used as index. Continue skips
				to the next stage of the k loop.*/
			}
			else // k>= 0
				rowi = k;//UPOS rowi=1
			for (i=0; rowi<horz_count; i++)
			{
				if (source_row[i] != rowbuffer[j][rowi])
				{
					stripe = 0; //Not a copy
					break; //Bail from this i loop
				}
				rowi++;
			} //UNEG i=1 now, UPOS i=horz_count-1 now.
			if (stripe) //If didn't turn stripe off yet, keep checking.
			{
				rowi = 0; //UNEG i=1 is right, UPOS
				for (; i<horz_count; i++)
				{
					if (source_row[i] != rowbuffer[j][rowi])
					{
						stripe = 0; //Not a copy
						break; //Bail from this i loop
					}
					rowi++;
				}
			}
			if (stripe) //The whole row matched; change the CA & bail.
			{
				if (type_ca != CA_REVERSIBLE && calist_ptr->stripeseedflag)
					Seed();
				if (calist_ptr->stripekillflag)
					Lambdalookup();
				else
				{
					if (calist_ptr->breedflag)	// Only keep score if breed cycle is enabled
						score += fail_stripe;	// penalty for fail stripe
				}
				if (calist_ptr->FocusCA() == this)
					update_flag = 1;					// update combo boxes
				return;
			}
		}
	}
}

/* This function takes a floating point and puts the decimal
part into decimal and the fractional part into fraction */
void Breakpoint(float number, int &decimal, int &fraction)
{
	decimal = number;	// Decimal
	number -= decimal;	// chop off decimal part
	number *= 100;	// bump fractional part of number over the decimal place
	fraction = number;	// Fractional
	number -= fraction;
	if (number >= .5)
		++fraction;	// Round up fraction
	if (fraction > 99)
	{
		fraction -= 100; ++fraction;	// Round up > .9, add to decimal
	}
	if (fraction < 0)
		fraction *= -1;	// Dont show the decimal part as negative!
}

/* This function takes two ints decimal and fraction and returns a
floating point number that is decimal.fraction */
float Makepoint(int decimal, int fraction)
{
	float number;
	number = decimal;
	number += fraction * 0.01;
	return (number);
}

void CA::Computeactual_lambda()
{
	unsigned short nonzero = 0;

	if (Gettype() != CA_STANDARD && Gettype() != CA_REVERSIBLE)
		return;

	for (unsigned short i = 0; i < nabeoptions; ++i)
		if (lookup[i])
			nonzero++;
	if (!nabeoptions)
		return;
	actual_lambda = (double)nonzero/(double)nabeoptions;
}

void CA::Symmetrize()
{
	unsigned short statesmask, sourcei, cell, mirrori;

	if (Gettype() != CA_STANDARD && Gettype() != CA_REVERSIBLE)
		return;

	statesmask = states-1;
	for (unsigned short i=0; i<nabeoptions; i++)
	{
		sourcei = i;
		mirrori = 0;
		for (unsigned int j=0; j<nabesize; j++)
		{
			mirrori <<= statebits;
			cell = statesmask & sourcei;
			sourcei >>= statebits;
			mirrori |= cell;
		}
//		assert(mirrori < nabeoptions);
		if (i < mirrori)
			lookup[mirrori] = lookup[i];
	}
}

void CA::Reverse(void)
{
	short i;
	unsigned char temp;

	if (type_ca != CA_REVERSIBLE)
		return;
	for (i = 0; i<horz_count; i++)
	{
		temp = past_row[i];
		past_row[i] = source_row[i];
		source_row[i] = temp;
	}
}

void CA::Settype(int newtype)
{
	unsigned char reseed = 1;

	if ( newtype != CA_STANDARD && newtype != CA_REVERSIBLE &&
		type_ca != CA_STANDARD && type_ca != CA_REVERSIBLE && type_ca != 0)
		reseed = 0;

	type_ca = newtype;
/*This weird usage is for doing member function pointers, and was
explained to rucker by Prof. Horstmann */
	switch(type_ca)
	{
		case CA_STANDARD:
			UpdateFunction = &CA::StandardUpdate;
			break;
		case CA_REVERSIBLE:
			UpdateFunction = &CA::ReversibleUpdate;
			break;
		case CA_WAVE:
			UpdateFunction = &CA::WaveUpdate;
			UpdateCell = &CA::WaveVelInt1;
			break;
		case CA_HEATWAVE:
//#define TWO_HEAT
#ifndef TWO_HEAT //look at one neighbor.
			UpdateFunction = &CA::WaveUpdate;
			UpdateCell = &CA::HeatInt1;
#else //look at two neighbors
			UpdateFunction = &CA::WaveUpdate2;
			UpdateCell2 = &CA::HeatInt2;
#endif //TWO_HEAT
			break;
		case CA_WAVE2:
			UpdateFunction = &CA::WaveUpdate2;
			UpdateCell2 = &CA::WaveVelInt2;
			break;
		case CA_OSCILLATOR:
			UpdateFunction = &CA::WaveUpdate;
			UpdateCell = &CA::Oscillator;
			break;
		case CA_OSCILLATOR_WAVE:
			UpdateFunction = &CA::WaveUpdate;
			UpdateCell = &CA::OscillatorWave;
			break;
		case CA_DIVERSE_OSCILLATOR:
			UpdateFunction = &CA::WaveUpdate;
			UpdateCell = &CA::DiverseOscillator;
			break;
		case CA_DIVERSE_OSCILLATOR_WAVE:
			UpdateFunction = &CA::WaveUpdate;
			UpdateCell = &CA::DiverseOscillatorWave;
			break;
		case CA_WATER_WAVE:
//Use for looking at bad space_step values.
			UpdateFunction = &CA::WaveUpdate;
			UpdateCell = &CA::UlamWave;
//			UpdateFunction = &CA::WaveUpdate2;
///			UpdateCell2 = &CA::WaterWave;
			break;
	}
	Changeradiusandstates();
	Adjust_acceleration_multiplier();
	if (reseed)
	{
		Seed();	// Need to seed with correct values & reposition row pointers!
		Setcolortable();
	}
}

inline void CA::Update(HDC hdc)
{
	(this->*UpdateFunction)(hdc);
}

void CA::Mutate(float mutation_strength)
{
	unsigned short i, bitstomutate;

	bitstomutate = nabeoptions * mutation_strength;
	if (!bitstomutate)
		bitstomutate = 1;
	if (nabeoptions)	// check for wave
		for (i=0; i<bitstomutate; i++)
			lookup[Random(nabeoptions)] ^= 1;
	Resetfreq();
	Computeactual_lambda();
//---BEGIN CA_wave
	space_step = 0.002 +
		space_step *(1.0 + (0.9*Randomsignreal()) * mutation_strength);
	time_step = 0.001 +
		time_step*(1.0 + (0.9*Randomsignreal()) * mutation_strength);
	max_intensity *= 1.0 + (0.9*Randomsignreal()) * mutation_strength;
	max_velocity *= 1.0 + (0.9*Randomsignreal()) * mutation_strength;
	friction_multiplier *= 1.0 + (0.9*Randomsignreal()) * mutation_strength;
	spring_multiplier *= 1.0 + (0.9*Randomsignreal()) * mutation_strength;
	mass *= 1.0 + (0.9*Randomsignreal()) * mutation_strength;
	driver_multiplier *= 1.0 + (0.9*Randomsignreal()) * mutation_strength;
	frequency_multiplier *= 1.0 + (0.9*Randomsignreal()) * mutation_strength;
	Adjust_acceleration_multiplier();
}

void CA::Change_time_step(int updown)
{
	if (fabs(time_step) < MIN_MULTIPLIER)
	{
		time_step *= -2.0;
		Adjust_acceleration_multiplier();
		return;
	}
	if (time_step < 0.0)
		updown *= -1;
	if (updown >0)
		time_step *= 1.1;
	else
		time_step *= 0.9;
	Adjust_acceleration_multiplier();
}

void CA::Change_space_step(int updown)
{
	if (fabs(space_step) < MIN_MULTIPLIER)
	{
		space_step *= -2.0;
		Adjust_acceleration_multiplier();
		return;
	}
	if (space_step < 0.0)
		updown *= -1;
	if (updown >0)
		space_step *= 1.1;
	else
		space_step *= 0.9;
	Adjust_acceleration_multiplier();
}

void CA::Change_spring(int updown)
{
	if (spring_multiplier == 0.0 && updown > 0)
		spring_multiplier = 0.01;
	if (updown >0)
		spring_multiplier *= 1.5;
	else
		spring_multiplier *= 0.6;
	if (spring_multiplier < 0.01)
		spring_multiplier = 0.0;
}

void CA::Change_mass(int updown)
{
	if (updown >0)
		mass *= 1.5;
	else
		mass *= 0.6;
	if (mass < START_MIN_MASS)
		mass = START_MIN_MASS;
}

void CA::Change_friction(int updown)
{
	if (friction_multiplier == 0.0 && updown > 0)
		friction_multiplier = MIN_MULTIPLIER;
	if (updown >0)
		friction_multiplier *= 1.5;
	else
		friction_multiplier *= 0.6;
	if (friction_multiplier < MIN_MULTIPLIER)
		friction_multiplier = 0.0;
}

void CA::Adjust_acceleration_multiplier()
{
	half_time_step = time_step/2.0;
	if (type_ca != CA_WATER_WAVE && type_ca != CA_HEATWAVE)
		space_step = time_step*2.0;
	acceleration_multiplier = time_step / (space_step * space_step);
	if (type_ca == CA_WAVE2)
		acceleration_multiplier /= 12.0;
//Because second diff is (-LL + 16L -30 C + 16R - RR) / (12*dx*dx)
//I figured this out in the Mathematica notebook wavenabe.ma by using
//Lagrangian interpolation to get a cubic through (-2,LL)(-1,L)(0,C)
//(1,R) and (2,RR), then taking the second derivative & evaluating at 0.
	frequency_factor = (2.0*PI*frequency_multiplier)/
	(time_step*OSCILLATOR_BASE_PERIOD);
   /* If frequency_multiplier is 1, this means that the oscillator
   has frequency_factor go from 0 to 2*pi as time goes from 0 to
   BASE_PERIOD*time_step.  If frequncy multiplier is 2 its twice as
   fast, if 0.5 half as fast, and so on.  I use a visually good
   BASE_PERIOD, like 200*/

}

void CA::Change_band_count(int updown)
{
	if (updown >0)
		band_count *= 2;
	else
		band_count /= 2;
	CLAMP(band_count, 2, MAX_COLOR);
	Setcolortable();
}

void CA::Change_max_intensity(int updown)
{
	Real step = 10;

	if (max_intensity > 100)
		step = 50;
	if (max_intensity < 5)
		step = 1;
	if (updown > 0)
		max_intensity += step;
	else
		max_intensity -= step;
	CLAMP(max_intensity , 1, 10000);
}

void CA::Change_heat_inc(int updown)
{
	if (heat_inc == 0.0 && updown > 0)
		heat_inc = MIN_MULTIPLIER;
	if (updown >0)
		heat_inc *= 1.5;
	else
		heat_inc *= 0.6;
	if (heat_inc < MIN_MULTIPLIER)
		heat_inc = 0.0;
	CLAMP(heat_inc, 0, 256);
}

void CA::Change_max_velocity(int updown)
{
	Real step = 10;

	if (max_velocity> 100)
		step = 50;
	if (max_velocity< 5)
		step = 1;
	if (updown > 0)
		max_velocity += step;
	else
		max_velocity -= step;
	CLAMP(max_velocity, 1.0, 1000.0);
}

void CA::Change_driver_amp(int updown)
{
	if (updown >0)
  {
  	if (!driver_multiplier)
     	driver_multiplier = 0.1;
     else
			driver_multiplier *= 1.5;
  }
	else
  {
		 driver_multiplier*= 0.6;
      if (driver_multiplier < 0.1)
      	driver_multiplier = 0.0;
  }
	CLAMP(driver_multiplier, 0.0, max_intensity);
}


void CA::Change_driver_freq(int updown)
{
	if (updown >0)
		frequency_multiplier *= 1.5;
	else
		 frequency_multiplier*= 0.6;
	CLAMP(frequency_multiplier, 0.1, 100);
  Adjust_acceleration_multiplier();
}

// ttl 3.12.95
void CA::Set_time_step(double ts)
{
	time_step = ts;
	if (fabs(time_step) < MIN_MULTIPLIER)
	{
		time_step *= -2.0;
		Adjust_acceleration_multiplier();
		return;
	}
	Adjust_acceleration_multiplier();
}

void CA::Set_space_step(double ss)
{
	space_step = ss;
	if (fabs(space_step) < MIN_MULTIPLIER)
	{
		space_step *= -2.0;
		Adjust_acceleration_multiplier();
		return;
	}
	Adjust_acceleration_multiplier();
}

void CA::Set_band_count(double bc)
{
	band_count = bc;
	CLAMP(band_count, 2, MAX_COLOR);
	Setcolortable();
}

void CA::Set_max_velocity(double mv)
{
	max_velocity = mv;
	CLAMP(max_velocity, 1.0, 1000.0);
}


void CA::Set_max_intensity(double mi)
{
	max_intensity = mi;
	CLAMP(max_intensity , 1, 10000);
}

void CA::Set_heat_inc(double hi)
{
	heat_inc = hi;
	if (heat_inc == 0.0)
		heat_inc = MIN_MULTIPLIER;
	if (heat_inc < MIN_MULTIPLIER)
		heat_inc = 0.0;
	CLAMP(heat_inc, 0, 256);
}

void CA::Set_spring(double s)
{
	spring_multiplier = s;
	if (spring_multiplier < START_MIN_SPRING)
		spring_multiplier = START_MIN_SPRING;
}

void CA::Set_friction(double f)
{
	friction_multiplier = f;
	CLAMP(friction_multiplier, 0.0, 100.0);
}

void CA::Set_driver_amp(double da)
{
	driver_multiplier = da;
	CLAMP(driver_multiplier, 1.0, max_intensity);
}

void CA::Set_driver_freq(double df)
{
	frequency_multiplier = df;
	CLAMP(frequency_multiplier, 1.0, 1000);
  Adjust_acceleration_multiplier();
}

void CA::Set_mass(double m)
{
	mass = m;
	if (mass < START_MIN_MASS)
		mass = START_MIN_MASS;
}

// tll

void CA::Setfailstripe(int updown)
{
	if (updown > 0)
		fail_stripe += 10;
	else
		fail_stripe -= 10;
}

void CA::Setentropybonus(int updown)
{
	if (updown > 0)
		entropy_bonus += 10;
	else
		entropy_bonus -= 10;
}

void CA::Setgeneratorflag(int onoff)
{
	generatorflag = onoff;
	if (generatorflag)
		Zeroseed();
}

//---------------------Start the CAlist things

CAlist::CAlist(HWND myhwnd, int maxcount)
{
	int i;
	count = START_CAS;
	mutation_strength = START_MUTATION_STRENGTH;
	hwnd = myhwnd;
	for (i=0; i<maxcount; i++)
	{
		list[i] = new CA(this);
		if (!list[i])
		{
			MessageBox(hwnd,
				(LPSTR)"Failure in CA Allocation!",
				(LPSTR)"Memory Problems!",
				MB_OK | MB_ICONEXCLAMATION );
			SendMessage(hwnd, WM_DESTROY, 0, 0L );
			return;
		}
	}
	focus = list[0];
	CA_count_per_edge = (int) sqrt(count);
	if ((CA_count_per_edge*CA_count_per_edge)<count)
		CA_count_per_edge++;
	sleep = 0;
	zoomflag = 0;
	breedflag = START_BREED_FLAG;
	breedcycle = START_BREED_CYCLE;
	breedcycle_count = 0;
	evolveflag = START_EVOLVEFLAG;
	mutateflag = START_MUTATE_FLAG;
	scrollflag = START_SCROLL_FLAG;
	stripekillflag = START_STRIPEKILL_FLAG;
	stripeseedflag = START_STRIPESEED_FLAG;
	viewmode = START_VIEWMODE;
	showvelocity = START_SHOWVELOCITY;
	wireflag = START_WIRE_FLAG;
	graphflag = START_GRAPH_FLAG;
}

CAlist::~CAlist()
{
	for (int i = 0; i < MAX_CAS; i++)
		delete list[i];
}

void CAlist::Locate(int dmaxx, int dmaxy)
{
//Don't do any Clear or Boxfocus in here, leave that to capow.cpp
	if (!zoomflag)
		for (int i=0; i<count; i++)
			list[i]->Locate(i, dmaxx, dmaxy, CA_count_per_edge);
	else
		focus->Locate(0, dmaxx, dmaxy, 1);
	blt_flag = 0;

}

void CAlist::Seed()
{
//#define UNIFORMSEED
#ifdef UNIFORMSEED
	unsigned char seedvalue;

	if (!zoomflag) //you can assume all horz_count same
		for (int j=0; j<GetSystemMetrics(SM_CXSCREEN); j++)
		{
			seedvalue = (unsigned char)Random(MAX_STATES); /* Use this in case there
	are different values of states in the different CAs*/
			for (int i=0; i<count; i++)
				list[i]->source_row[j] =
					seedvalue % (unsigned char)(list[i]->states);
		}
	else //when zoomflag, only seed the focus CA
		for (int j=0; j<GetSystemMetrics(SM_CXSCREEN); j++)
			focus->source_row[j] = (unsigned char)Random(focus->states);
#else //non UNIFORMSEED
	if (!zoomflag)
		for (short i=0; i<count; i++)
			list[i]->Seed();
	else
		focus->Seed();
#endif //UNIFORMSEED
}

void CAlist::Update_and_Show(HDC hdc)
{
	// computed_lines is initialized to zero in ca::locate to get in sync
	if (sleep)
		return;
	if (!zoomflag)	// not zoomed, do them all
	{
		for (int i=0; i<count; i++)
			list[i]->Update(hdc);
		blt_flag++;
		switch (viewmode)
		{
			case IDC_DOWN_VIEW:
//Minor bug here, this seems not to show lines at bottom of the box.
				if (!(blt_flag % BLT_LINES))
				{
#define GOACROSS
#ifndef GOACROSS

					if (list[0]->row_number == list[0]->miny)
					//In this case you just wrapped row_number to miny
						for (int i=0; i<count; i++)
						{  //blt the bottom strip
							WBM->WBMBitBlt(hdc,
								list[i]->minx,
								list[i]->maxy-BLT_LINES,
								list[i]->maxx,
								list[i]->maxy); //maxy is row_number - 1 here
						}
					else
						for (int i=0; i<count; i++)
						{
							WBM->WBMBitBlt(hdc,
								list[i]->minx,
								list[i]->row_number-BLT_LINES,
								list[i]->maxx,
								list[i]->row_number - 1);
						}
#else  //GOACROSS case does sqrt(count) BLTs instead of count BLTs : faster.
					int left_CA_index = 0;
					if (list[0]->row_number == list[0]->miny)
					//In this case you just wrapped row_number to miny
						for (int i=0; i<CA_count_per_edge; i++)
						{
							WBM->WBMBitBlt(hdc,
								list[left_CA_index]->minx,
								list[left_CA_index]->maxy-BLT_LINES,
								list[left_CA_index + CA_count_per_edge - 1]->maxx,
								list[left_CA_index]->maxy);
							left_CA_index += CA_count_per_edge;
						}
					else
						for (int i=0; i<CA_count_per_edge; i++)
						{
							WBM->WBMBitBlt(hdc,
								list[left_CA_index]->minx,
								list[left_CA_index]->row_number-BLT_LINES,
								list[left_CA_index + CA_count_per_edge - 1]->maxx,
								list[left_CA_index]->row_number-1);
							left_CA_index += CA_count_per_edge;
						}
#endif //GOACROSS
					blt_flag = 0;
				}
				break;
			case IDC_SCROLL_VIEW:
				if (!(blt_flag % BLT_LINES))
				{
					WBM->WBMBitBlt(hdc, list[0]->minx, list[0]->miny,
							list[count-1]->maxx, list[count-1]->maxy);
					blt_flag = 0;
				}
				break;
			case IDC_WIRE_VIEW:
				break;
			case IDC_GRAPH_VIEW:
				WBM->WBMBitBlt(hdc, list[0]->minx, list[0]->miny,
							list[count-1]->maxx, list[count-1]->maxy);
				break;
		}
		if (breedflag)  //only breed in non zoomflag mode
		{
			++breedcycle_count;
			if (breedcycle_count >= breedcycle)
			{
				Evolve();	// Evolving methods for breeding
				breedcycle_count = 0;
			}
		}
	}// end non zoomflag  case
	else
	{
		focus->Update(hdc);	// We are zoomed, just do the focused one
		blt_flag++;
		switch (viewmode)
		{
			case IDC_GRAPH_VIEW:
				WBM->WBMBitBlt(hdc, focus->minx, focus->miny,
							focus->maxx, focus->maxy);
				break;
			case IDC_DOWN_VIEW:
				if (!(blt_flag % BLT_LINES))
				{
					if (focus->row_number == focus->miny)
					//In this case you just wrapped row_number to miny
					//blt the bottom strip
						WBM->WBMBitBlt(hdc,
							focus->minx,
							focus->maxy-BLT_LINES,
							focus->maxx,
							focus->maxy); //maxy is row_number - 1 here
					else
						WBM->WBMBitBlt(hdc,
							focus->minx,
							focus->row_number-BLT_LINES,
							focus->maxx,
							focus->row_number - 1);
					blt_flag = 0;
				}
				break;
			case IDC_SCROLL_VIEW:
				if (!(blt_flag % BLT_LINES))
				{
					WBM->WBMBitBlt(hdc, focus->minx, focus->miny,
							focus->maxx, focus->maxy);
					blt_flag = 0;
				}
				break;
			case IDC_WIRE_VIEW:
				break;
		}
	} // end  zoomflag  case
	// Check to see if it is Breeding time
}

void CAlist::Evolve()
{//This is called from update_and_show every breeding cycle
	int field,i, mother, father;
	CA* oldfocus;

	if (zoomflag) //Don't do anything while zoom is on.
	{
		Resetscores();	// reset all scores
		return;
	}
	Sortfitness();	/* get a list of the best to the worst ca's
		This includes a permute, so no fixed order advantage.*/
	field = count/4;
	/* 1/4th get left alone, 1/2th stay the same, 1/4th get altered.
		Also the very worst guy will be zapped.  Use focus as index.*/
	oldfocus = focus;
	for (i = field*3; i < count-1; i++)
	{
		focus = list[best_list[i]];
		switch (evolveflag)
		{
			case ZAP:
			/* randomize yourself */
				focus->Lambdalookup();
				break;
			case CLONE:
			/* copy a winner */
				focus->CopyCA(
					list[best_list[Random(field)]]);
				break;
			case SEX:
			/* be a child of two winners */
				mother = Random(field);
				father = Random(field);
				if (field > 1)	// more than one winner to be parents?
					while (mother == father)
						// dont want mother to be the father also
						father = Random(field);
				focus->Become_child_of(
					list[best_list[mother]], list[father]);
				break;
			case NONE:
				break;
		}
	}
	focus = oldfocus;	// reset focus to normal
	/* Always zap the biggest loser */
	list[best_list[count-1]]->Lambdalookup();
/*Mutate everyone if that's in the cards*/
	if (mutateflag)
		for (i = 0; i < count; i++)
			list[i]->Mutate(mutation_strength);
	Resetscores();	// reset all scores
	MessageBeep(0);	// Beep!
}

void CAlist::Setviewmode(int newmode)
{
	int i;
	viewmode = newmode;
	switch (viewmode)
	{
		case IDC_DOWN_VIEW :
			for (i = 0; i<count; i++)
				list[i]->row_number = list[i]->miny;
			break;
		case IDC_SCROLL_VIEW :
			if (!zoomflag)
				for (i = 0; i<count; i++)
					WBM->ClearSection(list[i]->minx, list[i]->miny,
						list[i]->maxx, list[i]->maxy);
			else
				WBM->ClearSection(focus->minx, focus->miny,
					focus->maxx, focus->maxy);
			for (i = 0; i<count; i++)
				list[i]->row_number = list[i]->maxy - BLT_LINES + 1;
			break;
		case IDC_GRAPH_VIEW:
			if (!zoomflag)
				for (i = 0; i<count; i++)
					WBM->ClearSection(list[i]->minx, list[i]->miny,
						list[i]->maxx, list[i]->maxy);
			else
				WBM->ClearSection(focus->minx, focus->miny,
					focus->maxx, focus->maxy);
			break;
		default : ;
	}
}

void CAlist::Setentropy(int onoff)
{
	for (int i=0; i<count; i++)
		list[i]->Setentropyflag(onoff); //Calls Resetfreq

}

void CAlist::Setwrapflag(int onoff)
{

	for (int i = 0; i < count; i++)
		list[i]->Setwrapflag(onoff);
}

void CAlist::Setgeneratorflag(int onoff)
{

	for (int i = 0; i < count; i++)
		list[i]->Setgeneratorflag(onoff);
}

void CAlist::Boxfocus(HDC hdc, COLORREF color)
{
	WBM->PutRectangle(hdc, focus->minx-1, focus->miny-1,
			focus->maxx+1, focus->maxy+1, color);
}

int CAlist::Setfocus(HDC hdc, CA *new_focus)
{
	if (zoomflag)
		return 0;
	if (new_focus==focus)
		return 1;
	Boxfocus(hdc,RGB(0,0,0));
	focus = new_focus;
	Boxfocus(hdc,RGB(255,255,255));
	return 0;
}

CA *CAlist::Getfocus(int x, int y)
{
/* I do minx - BORDER instead of minx so that the cursor won't be in a dead
zone all along the left edge of the screen.  Use "BORDER" because in Locate
below, the left and right edges are offset from disp_ptr->minx by 1. So
you would think I could use "1" here instead of "BORDER", but that doesn't
work so what the hey use BORDER which happened to be 3 when I wrote this
comment at calife1.zip.*/
	for (int i=0; i<count; i++)
		if (list[i]->minx - BORDER <= x && x <= list[i]->maxx + BORDER &&
			list[i]->miny - BORDER  <= y && y <= list[i]->maxy + BORDER)
				return list[i];
	return focus;
}

int CAlist::Bumpcount(int updown)
{ //Return 0 if no change, else return 1.
	return Changecount((CA_count_per_edge+updown) * (CA_count_per_edge+updown));
}

int CAlist::Changecount(int icount)
{
 //Return 0 if no change happens, else return 1.

	int i;
	RECT rect;
	int oldcount;

	if (icount > MAX_CAS)
		icount = MAX_CAS;
	if (icount < 1)
		icount = 1;
	if (count == icount)
		return 0;

	oldcount = count;
	count = icount;
	CA_count_per_edge = sqrt(count);
 //	assert (CA_count_per_edge*CA_count_per_edge == count);
//	if ((CA_count_per_edge*CA_count_per_edge)<count)
  //		CA_count_per_edge++;
	GetClientRect(hwnd, &rect); //Here is where I emulate WM_SIZE
	Locate(rect.right,rect.bottom);//for possible new CAs.
	if (count > oldcount)	// Expand field
		for (i=oldcount; i< count; i++)
		{
			list[i]->Changeradiusandstates(focus->radius, focus->states);
			list[i]->Settype(focus->Gettype());
		}
	else	// Shrink field
		for (i = count; i < oldcount; i++)
			if (focus == list[i])	// if focus being deleted
				focus = list[0];		// set to the first CA

	SendMessage(hwnd, WM_COMMAND, IDM_CLEAR, 0L);
	update_flag = 1;	// need to update dialog boxes
	return 1;
}

int CAlist::Zoom(int onoff)
{
	if (onoff == zoomflag)
		return 0;
	zoomflag = onoff;
	return 1;
}

void CAlist::DrawDivider(HDC hdc)
{
	if (!zoomflag)
		for (int i = 0; i<count; i++)
			WBM->PutRectangle(hdc, list[i]->minx-3,
				list[i]->miny-3, list[i]->maxx+3, list[i]->maxy+3,
				RGB(128,128,128));
}

void CAlist::Sortfitness()
{
#define RUDYWAY
#ifndef RUDYWAY
	int i, j;
	int used[MAX_CAS];	/* keeps track of which ca's have been recorded
									in the best list */

	for (i = 0; i < count; i++)
		used[i] = 0;

	for (j = 0; j < count; j++)
	{
		best_list[j] = 0;
		while (used[best_list[j]])	// initialize it with the 1st unclaimed ca
			++best_list[j];

		for (i=0; i<count; i++)	/* now test it with the rest of the unclaimed
											ca's to see which is the next best */
			if (used[i] == 0)
			{
				if (list[i]->score > list[best_list[j]]->score)
					best_list[j] = i;
				else if (list[i]->score == list[best_list[j]]->score &&
					Random(2)) //To equalize chances of winning.
					best_list[j] = i;
			}
		used[best_list[j]] = 1;	// show this ca as claimed
	}
#else
	int i, j, temp;

	jolt_permute_buffer(count); //This function lives in random.cpp
	for (i = 0; i < count; i++)
		best_list[i] = permute_buffer[i]; //buffer is in random.h
	/* permute the order and then do a bubble sort */

	for (j = 0; j < count-1; j++) //Fix the first member, then next, etc.
		for (i=j+1; i<count; i++)
			if (list[best_list[i]]->score > list[best_list[j]]->score)
			{ //swap
				temp = best_list[j];
				best_list[j] = best_list[i];
				best_list[i] = temp;
			}
#endif
}

//---Commands---------------------------------------------------

void CAlist::Changeradiusandstates(int iradius, int istates)
{
	for (int i = 0; i < count; i++)
		list[i]->Changeradiusandstates(iradius, istates);
}

void CAlist::Symmetrize()
{
	for (int i = 0; i < count; i++)
		list[i]->Symmetrize();
}

void CAlist::Reverse()
{
	for (int i = 0; i < count; i++)
		list[i]->Reverse();
}

void CAlist::Oneseed()
{
	for (int i = 0; i < count; i++)
		list[i]->Oneseed();
}
void CAlist::Zeroseed()
{
	for (int i = 0; i < count; i++)
		list[i]->Zeroseed();
}
void CAlist::Sineseed()
{
	for (int i = 0; i < count; i++)
		list[i]->Sineseed();
}

void CAlist::Setcolortable()
{
	for (int i = 0; i < count; i++)
		list[i]->Setcolortable();
}

void CA::Setlambda(float newlambda)
{
	unsigned short i;
  float increment;
	float oldlambda, flip_prob;

	oldlambda = lambda;
	lambda = newlambda;
	if (lambda > MAX_LAMBDA)
		lambda = MAX_LAMBDA;
	if (lambda < MIN_LAMBDA)
		lambda = MIN_LAMBDA;
  increment = lambda - oldlambda;

	if (increment > 0) //increase lambda
	{
		flip_prob = (lambda - oldlambda)/(1.0 - oldlambda);
/* Because lambda = ((1.0 - oldlambda)*flip_prob*nabeoptions+
oldlambda*nabeoptions) / nabeoptions; or
lambda = (1-oldlambda)*flip_prob + oldlambda.*/
		for (i=0; i<nabeoptions; i++)
			if (!lookup[i] && Randomreal()< flip_prob )
				lookup[i] = 1 + Random(states - 1);
	}
	else //decrease lambda
	{
		flip_prob = (oldlambda - lambda)/(oldlambda);
/* Because lambda = (oldlambda*nabeoptions -
oldlambda*nabeoptions*flip_prob) / nabeoptions; or
lambda = oldlambda - oldlamda*flip_prob.*/
		for (i=0; i<nabeoptions; i++)
			if (lookup[i] && Randomreal()< flip_prob )
				lookup[i] = 0;
	}
	Computeactual_lambda();
	Resetfreq();
}

void CAlist::Changelambda(float newlambda)
{
	for (int i = 0; i < count; i++)
		list[i]->Changelambda(newlambda);
}

void CAlist::Changeentropy(double newentropy)
{
	for (int i = 0; i < count; i++)
		list[i]->Changeentropy(newentropy);
}

void CAlist::Lambdalookup()
{
	for (int i = 0; i < count; i++)
	{
		list[i]->Lambdalookup();
	}
}

void CAlist::Mutate()
{
	for (int i = 0; i < count; i++)
		list[i]->Mutate(mutation_strength);
}

void CAlist::Breed()
{
	if (count == 1 || focus == list[0])
		return;	// breeding mom with mom does nothing
	for (int i = 1; i < count; i++)
		if (list[i] != focus)
			list[i]->Become_child_of(focus, list[0]);
	Seed();
}

void CAlist::Copymutate()
{
	for (int i = 0; i < count; i++)
		if (list[i] != focus)
		{
			list[i]->Settype(focus->Gettype());
			list[i]->CopyCA(focus);
			list[i]->Mutate(mutation_strength);
		}
}

void CAlist::SetWindowBitmap(WindowBitmap *myWBM)
{
	WBM = myWBM;	// To put the box focus into the bitmap
	for (int i = 0; i < MAX_CAS; i++)
		list[i]->WBM = myWBM;
}

void CAlist::SetSleep(int onoff)
{
	sleep = onoff;
}

void CAlist::SetCAType(CA *target, unsigned char newtype)
{
	target->Settype(newtype);
}

void CAlist::SetAllType(unsigned char newtype)
{
	CA *catarget;
	for (int i = 0; i<count; i++)
	{
		catarget = list[i];
		SetCAType(catarget, newtype);
	}
	SendMessage(hwnd, WM_COMMAND, IDM_CLEAR, 0L);
}


void CAlist::Setshowmode(int newmode)
{
	for (int i = 0; i < count; i++)
		list[i]->Setshowmode(newmode);
}

void CAlist::Change_time_step(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Change_time_step(updown);
}

void CAlist::Change_space_step(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Change_space_step(updown);
}

void CAlist::Change_spring(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Change_spring(updown);
}
void CAlist::Change_mass(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Change_mass(updown);
}

void CAlist::Change_friction(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Change_friction(updown);
}

void CAlist::Change_driver_amp(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Change_driver_amp(updown);
}

void CAlist::Change_driver_freq(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Change_driver_freq(updown);
}

void CAlist::Change_band_count(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Change_band_count(updown);
}

void CAlist::Change_max_intensity(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Change_max_intensity(updown);
}
void CAlist::Change_heat_inc(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Change_heat_inc(updown);
}

void CAlist::Change_max_velocity(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Change_max_velocity(updown);
}

void CAlist::Setfailstripe(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Setfailstripe(updown);
}

void CAlist::Setentropybonus(int updown)
{
	for (int i = 0; i < count; i++)
		list[i]->Setentropybonus(updown);
}

void CAlist::Setbreedcycle(int updown)
{
	if (updown > 0)
		breedcycle += 50;
	else
		breedcycle -= 50;
	CLAMP(breedcycle, 50, 10000);
}

void CAlist::Setmutation(int updown)
{
	float	increment;
  if (mutation_strength >= 0.2)
  	increment = 0.1;
  else if (mutation_strength >= 0.1)
	increment = 0.01;
	if (updown > 0)
		mutation_strength += increment;
	else
		mutation_strength -= increment;
	CLAMP(mutation_strength, 0.001, 0.90);
}

void CAlist::Resetscores()
{
	for (int i = 0; i < count; i++)
		list[i]->Resetscore();
}

void CAlist::Randomize()
{
	unsigned char type;
	int stateradiusindex;

	for (int i = 0; i < count; i++)
	{
		if (Random(2))
			type = CA_STANDARD;
		else
			type = Random(MAX_TYPE_CA)+CA_STANDARD;
		SetCAType(list[i], type);
		if (type == CA_STANDARD || type == CA_REVERSIBLE)
		{
			stateradiusindex =  Random(STATERADIUSCOUNT);
			list[i]->Changeradiusandstates(StartRadius[stateradiusindex],
				StartStates[stateradiusindex]);
		}
		list[i]->Resetscore();
	}
}

void CAlist::CopyCA(CA *ca1, CA *ca2)
{
	if (ca1->Gettype() != ca2->Gettype())
		SetCAType(ca2, ca1->Gettype());
	ca2->CopyCA(ca1);
}

void CAlist::Saveindividual(char *filename, CA *target)
{
	ofstream ofs;

	ofs.open(filename, ios::noreplace);
	if (ofs.fail())
	{ // file filename already exists.  Hit this code if yes.
		if (MessageBox( hwnd,
			(LPSTR)"Do you want to overwrite the existing file?",
			(LPSTR)"",
			MB_YESNO | MB_ICONEXCLAMATION ) == IDNO)
		{
			ofs.close();
			return;
		}
	}

	ofs.close();   // close previous open
	ofs.open(filename, ios::out);
	if (ofs.fail())
	{  // cannot open file filename for output
		MessageBox( hwnd,	(LPSTR)"File Creation failure!",
			(LPSTR)"Disk Problems!",
			MB_OK | MB_ICONEXCLAMATION );
		ofs.close();
		return;
	}

	ofs << target;
	ofs.close();
}

void CAlist::Saveall(char *filename)
{
	ofstream ofs;

	ofs.open(filename, ios::noreplace);
	if (ofs.fail())
	{ // file filename already exists.  Hit this code if yes.
		if (MessageBox( hwnd,
			(LPSTR)"Do you want to overwrite the existing file?",
			(LPSTR)"",
			MB_YESNO | MB_ICONEXCLAMATION ) == IDNO)
		{
			ofs.close();
			return;
		}
	}

	ofs.close();   // close previous open
	ofs.open(filename, ios::out);
	if (ofs.fail())
	{  // cannot open file filename for output
		MessageBox( hwnd,	(LPSTR)"File Creation failure!",
			(LPSTR)"Disk Problems!",
			MB_OK | MB_ICONEXCLAMATION );
		ofs.close();
		return;
	}

	ofs << endl << "Number of CAs is " << count;
	for (int i = 0; i < count; i++)
		ofs << list[i];

//	ofs << endl << "Zoom flag is " << (unsigned int) zoomflag;
	ofs << endl << "Mutation strength is " << mutation_strength;
	ofs << endl << "CA count per edge is " << CA_count_per_edge;
	ofs << endl << "Mutate flag is " << (unsigned int) mutateflag;
	ofs << endl << "Breed flag is " << (unsigned int) breedflag;
	ofs << endl << "Breed cycle is " << breedcycle;
	ofs << endl << "Bread cycle count is " << breedcycle_count;
	ofs << endl << "Evolve flag is " << (unsigned int) evolveflag;
	ofs << endl << "View mode is " << viewmode;
	ofs << endl << "Show velocity is " << showvelocity;
	ofs << endl << "Graph flag is " << (unsigned int) graphflag;
	ofs << endl << "Wire flag is " << (unsigned int) wireflag;
	ofs << endl << "Scroll flag is " << (unsigned int) scrollflag;
	ofs << endl << "Stripekill flag is " << (unsigned int) stripekillflag;
	ofs << endl << "Stripe seed flag is " << (unsigned int) stripeseedflag;

	ofs.close();
}
void CAlist::Loadindividual(char* filename, CA *target)
{
	ifstream ifs(filename, ios::in);

	if (ifs.fail())
	{
		MessageBox( hwnd,
			(LPSTR)"Can't open file!",
			(LPSTR)"File Problems!",
			MB_OK | MB_ICONEXCLAMATION );
		ifs.close();
		return;
	}

	ifs >> target;
	ifs.close();

/* If we load a new file, we may want to start breed cycle all over to
	give the new ca a fighting chance */
	if (breedflag)
 /*		if (MessageBox( hwnd,
			(LPSTR)"Reset Breedcycle counter and all scores?",
			(LPSTR)"",
			MB_YESNO | MB_ICONEXCLAMATION ) == IDYES)   */
		{
			Resetscores();
			breedcycle_count = 0;
		}
}

void CAlist::Loadall(char* filename)
{
	char dummybuf[80];
	unsigned int uiread;

	ifstream ifs(filename, ios::in);

	if (ifs.fail())
	{
		MessageBox( hwnd,
			(LPSTR)"Can't open file!",
			(LPSTR)"File Problems!",
			MB_OK | MB_ICONEXCLAMATION );
		ifs.close();
		return;
	}

	// "\n Number of CAs is %d"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> dummybuf >> count;
	for (int i = 0; i < count; i++)
		ifs >> list[i];

	// "\n Zoom flag is %"
//	ifs.get();  // read off eoln
//	ifs >> dummybuf >> dummybuf >> dummybuf >>  uiread;
 //	zoomflag = (unsigned char) uiread;
	// "\n Mutation strength is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> mutation_strength;
	// "\n CA count per edge is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> dummybuf >> dummybuf
		 >> CA_count_per_edge;
	// "\n Mutate flag is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> uiread;
	mutateflag = (unsigned char) uiread;
	// "\n Breed flag is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> uiread;
	breedflag = (unsigned char) uiread;
	// "\n Breed cycle is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> breedcycle;
	// "\n Bread cycle count is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> dummybuf
		 >> breedcycle_count;
	// "\n Evolve flag is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> uiread;
	evolveflag = (unsigned char) uiread;
	// "\n View mode is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> viewmode;
	// "\n Show velocity is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> showvelocity;
	// "\n Graph flag is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> uiread;
	graphflag = (unsigned char) uiread;
	// "\n Wire flag is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> uiread;
	wireflag = (unsigned char) uiread;
	// "\n Scroll flag is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> uiread;
	scrollflag = (unsigned char) uiread;
	// "\n Stripekill flag is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> uiread;
	stripekillflag = (unsigned char) uiread;
	// "\n Stripe seed flag is %"
	ifs.get();  // read off eoln
	ifs >> dummybuf >> dummybuf >> dummybuf >> dummybuf >> uiread;
	stripeseedflag = (unsigned char) uiread;

	ifs.close();

/* If we load a new file, we may want to start breed cycle all over to
	give the new ca a fighting chance */
	if (breedflag)
/*		if (MessageBox( hwnd,
			(LPSTR)"Reset Breedcycle counter and all scores?",
			(LPSTR)"",
			MB_YESNO | MB_ICONEXCLAMATION ) == IDYES)  */
		{
			Resetscores();
			breedcycle_count = 0;
		}

/* trigger a window resizing event to
	clear and redraw the display*/
	SendMessage(hwnd, WM_SIZE, NULL, 0L);
}
//--------------------Generator things-----------------

Generator::Generator(CA *myca)
{
	ca_ptr = myca;
	intensity = 0.0;
	velocity = 0.0;
	location =1; //Put it one step from the edge.
	state = 0;
	time = 0.0;
}

 void Generator::Step()
{
/* This is called right after the target row is updated, but before
it is shown to the bitmap and swapped with the source row.  This way,
the generator value gets in as the visible target value, and is used
as the new source value.  We use the internal onoff flag to decide
whether to really do this, that way we can always call this Step.*/
	intensity = sin(time);
	velocity = -cos(time);
	/* Set the wave value to generator intensity times one eighth of
		the max possible. */
	if (ca_ptr->Gettype() != CA_STANDARD &&
		ca_ptr->Gettype() != CA_REVERSIBLE)
	{
		ca_ptr->wave_target_row[location].intensity =
			intensity * (ca_ptr->max_intensity / 8.0);//MAX_GENERATOR_INTENSITY;
		ca_ptr->wave_target_row[location].velocity = 0; /*You don't
			use this veloicty, and nobody else looks at it.*/
/*		if (ca_ptr->Gettype() == CA_WAVE_CORRECT2)
		{ //Do a second point so big nabes notice.
		ca_ptr->wave_target_row[location+1].intensity =
			intensity * MAX_GENERATOR_INTENSITY;//(ca_ptr->max_intensity / 2.0);
		ca_ptr->wave_target_row[location+1].velocity = 0.0;
		}
 */
	}
	else
	{
	//Set the standard value in the range 0 to maximum value (states - 1)
	//Note that this osicallates with half the time of period, due to fabs.
		state = (ca_ptr->states) * fabs(intensity);
		if (state == ca_ptr->states)
			state--; //Just in case you hit an overflow via intensity 1.0
		ca_ptr->target_row[location] = state;
	}
//Now step the time.
	time += ca_ptr->time_step;
	if (time > 2.0*PI)
		time -= 2.0*PI;
}


// tl 10.29.94

void CA::Install_Ramp(int paramcode, double rampvalue)
{
	switch (paramcode)
	{
		case RAMP_ACCEL:
			Set_space_step(rampvalue);
			break;

		case RAMP_VEL_MULT:
			Set_time_step(rampvalue);
			break;

		case RAMP_MAX_VEL:
			Set_max_velocity(rampvalue);
			break;

		case RAMP_LAMBDA:
			Setlambda(rampvalue);
			break;

		case RAMP_BAND:
			Set_band_count(rampvalue);
			break;

		case RAMP_MAX_INTENSITY:
			Set_max_intensity(rampvalue);
			break;

		case RAMP_HEAT_COUNT:
			Set_heat_inc(rampvalue);
			break;

		case RAMP_FRICTION:
			Set_friction(rampvalue);
			break;

		case RAMP_SPRING:
			Set_spring(rampvalue);
			break;

		case  RAMP_DRIVER_AMP:
			Set_driver_amp(rampvalue);
			break;

		case  RAMP_DRIVER_FREQ:
			Set_driver_freq(rampvalue);
			break;

		case RAMP_INDUCTANCE:
			Set_mass(rampvalue);
			break;
	}
}


//  tl 10.29.94

void CAlist::Install_Ramp(int paramcode, double ramptable[])
{
	for (int i = 0; i < count; i++)
		list[i]->Install_Ramp(paramcode, ramptable[i]);
}

// tl 10-31-94

void CAlist::CheckRange (int paramcode, double minmax[])
{
	double tmp;
	if (minmax[0] > minmax[1])
		{
			tmp = minmax[0];
			minmax[0] = minmax[1];
			minmax[1] = tmp;
		}

	switch (paramcode)
	{
		case RAMP_ACCEL:
		case RAMP_VEL_MULT:
			if (minmax[0] < MIN_MULTIPLIER)
				minmax[0] = MIN_MULTIPLIER;
			if (minmax[1] < minmax[0])
				minmax[1] = minmax[0];
			break;

		case RAMP_LAMBDA:
			CLAMP (minmax[0], MIN_LAMBDA, MAX_LAMBDA);
			CLAMP (minmax[1], MIN_LAMBDA, MAX_LAMBDA);
			break;

		case RAMP_BAND:
			CLAMP (minmax[0], 2, MAX_COLOR);
			CLAMP (minmax[1], 2, MAX_COLOR);
			break;

		case RAMP_MAX_VEL:
			CLAMP(minmax[0], 1.0, 1000.0);
			CLAMP(minmax[1], 1.0, 1000.0);
			break;

		case RAMP_MAX_INTENSITY:
			CLAMP(minmax[0], 1, 10000);
			CLAMP(minmax[1], 1, 10000);
			break;

		case RAMP_HEAT_COUNT:
			CLAMP(minmax[0], 0, 256);
			CLAMP(minmax[1], 0, 256);
			break;

		case RAMP_FRICTION:
			if (minmax[0] < MIN_MULTIPLIER)
				minmax[0] = MIN_MULTIPLIER;
			if (minmax[1] < minmax[0])
				minmax[1] = minmax[0];
			break;

		case RAMP_SPRING:
			if (minmax[0] < START_MIN_SPRING)
				minmax[0] = START_MIN_SPRING;
			if (minmax[1] < minmax[0])
				minmax[1] = minmax[0];
			break;

		case  RAMP_DRIVER_AMP:
			CLAMP(minmax[0], 1.0, 10000 / 2.0);
			CLAMP(minmax[1], 1.0, 10000 / 2.0);
			break;

		case  RAMP_DRIVER_FREQ:
			CLAMP(minmax[0], 1.0, 1000);
			CLAMP(minmax[1], 1.0, 1000);
			break;

		case RAMP_INDUCTANCE:
			if (minmax[0] < START_MIN_MASS)
				minmax[0] = START_MIN_MASS;
			if (minmax[1] < minmax[0])
				minmax[1] = minmax[0];
			break;
	}
}

void CA::RealWrap(Real &vx, Real vlo, Real vhi)
 //  {if (vx<vlo){vx = vlo;return;} if (vx>vhi)vx = vhi;return;}
{   //assyne vl0 = -vhi, and vhi > 0.
	if (vx<vlo)
	{
		while (vx < vlo + vlo)
			vx += vhi;
		vx = vhi - (vlo-vx);
		return;
	}
	if (vx>vhi)
	{
		while (vx > vhi + vhi)
			vx -= vhi;
		vx = vlo + (vx - vhi);
	}
	return;
}


ofstream& operator<<(ofstream& ofs, CA* target)
{
	short i,j;

	ofs << endl << "1D CA file format " << FILEVERSION << '.';
	ofs << endl << "CA type " << (unsigned int) target->type_ca;
	ofs << endl << "radius " << target->radius
		 << " states " << target->states << '.';

	/* We don't write the lookup table to the file if we are dealing
		with a Wave or Heatwave CA */
	if ((unsigned int) target->type_ca >= CA_WAVE)
		ofs << endl << "Lookup table is not used." << endl;
	else
	{
		ofs << endl << "The lookup table with "
			 << target->nabeoptions << " entries follows:" << endl;
			/*This should be %ud so 32K doesn't come out
			negative, but if I change it now, my existing *.ca files
			become unreadable! I need a temporary alternate load version 1
			and load version 2 to fix */
		j = 0;
		for (i = 0; i < target->nabeoptions; i++)
		{
			ofs << (unsigned int) target->lookup[i] << ',';
			if (++j>26)
				{ ofs << endl;
				  j = 0;
				}
		}
	}

	ofs << endl << "The color table with " << target->states
		 << " Red,Green,Blue values follows:" << endl;
	j = 0;
	for (i = 0; i < target->states; i++)
	{
		ofs << (unsigned int) GetRValue(target->colortable[i]) << ','
			 << (unsigned int) GetGValue(target->colortable[i]) << ','
			 << (unsigned int) GetBValue(target->colortable[i]) << ',';
		if (++j>6)
			{	ofs << endl;
				j = 0;
			}
	}


	ofs << endl << "Show mode is " << target->showmode;
	ofs << endl << "Generator flag is " << (unsigned int) target->generatorflag;
	ofs << endl << "Wrap flag is " << (unsigned int) target->wrapflag;
	// print floats in fixed point
	ofs << setiosflags(ios::fixed) << setprecision(2);
	ofs << endl << "Lambda value of " << target->lambda;
	ofs << endl << "Entropy level of " << target->target_entropy;
	ofs << endl << "Entropy score of " << target->entropy_bonus;
	ofs << endl << "Failstripe value of " << target->fail_stripe;

	/* move all rows back to default start position, but retain all the current
	information in them, so we save a whole line of valid information */
	target->SyncRows();

	switch ((unsigned int) target->type_ca)
	{
		case CA_STANDARD:
			ofs << endl << "The source_row with "
				 << GetSystemMetrics(SM_CXSCREEN)
				 << " values follows:" << endl;
			j = 0;
			for (i = 0; i < GetSystemMetrics(SM_CXSCREEN); i++)
			{
				ofs << (unsigned int) target->source_row[i] << ',';
				if (++j>26)
					{	ofs << endl;
						j = 0;
					}
			}
			break;

		case CA_REVERSIBLE:
			ofs << endl << "The source_row with "
				 << GetSystemMetrics(SM_CXSCREEN)
				 << " values follows:" << endl;
			j = 0;
			for (i = 0; i < GetSystemMetrics(SM_CXSCREEN); i++)
			{
				ofs << (unsigned int) target->source_row[i] << ',';
				if (++j>26)
					{	ofs << endl;
						j = 0;
					}
			}

			j = 0;
			ofs << endl << "The past_row with "
				 << GetSystemMetrics(SM_CXSCREEN)
				 << " values follows:" << endl;
			for (i = 0; i < GetSystemMetrics(SM_CXSCREEN); i++)
			{
				ofs << (unsigned int) target->past_row[i] << ',';
				if (++j>26)
					{	ofs << endl;
						j = 0;
					}
			}
			break;

		case CA_WAVE:
		case CA_HEATWAVE:
		case CA_WAVE2:
		case CA_OSCILLATOR:
		case CA_OSCILLATOR_WAVE:
		case CA_DIVERSE_OSCILLATOR:
		case CA_DIVERSE_OSCILLATOR_WAVE:
		case CA_WATER_WAVE:
			ofs << endl << "The wave_source_row with "
				 << GetSystemMetrics(SM_CXSCREEN)
				 << " values follows:" << endl;
			j = 0;
			for (i = 0; i < GetSystemMetrics(SM_CXSCREEN); i++)
			{
			/* intensity and velocity could be negative */
				ofs << target->wave_source_row[i].intensity << ',';
				ofs << target->wave_source_row[i].velocity << ',';
				if (++j>8)
				{	ofs << endl;
					j = 0;
				}
			}
			ofs << endl << "Time step is " << target->time_step;
			ofs << endl << "Space step is " << target->space_step;
			ofs << endl << "Band count is " << target->band_count;
			ofs << endl << "Heat increment is " << target->heat_inc;
			ofs << endl << "Intensity range is " << target->max_intensity;
			ofs << endl << "Velocity range is " << target->max_velocity;
			ofs << endl << "Friction multiplier is " << target->friction_multiplier;
			ofs << endl << "Spring multiplier is " << target->spring_multiplier;
			ofs << endl << "Driver multiplier is " << target->driver_multiplier;
			ofs << endl << "Frequency multiplier is " << target->frequency_multiplier;
			ofs << endl << "Mass is " << target->mass;
			break;
	}
	return ofs;
}


ifstream& operator>>(ifstream& ifs, CA* target)
{
	char dummybuf[80];
	char ch;
	short i, j;
	unsigned int uiread,red,green,blue;
//	int dec,fra;
	Real dummyreal;
	int version, istates, iradius, itype, iwidth, readwidth;

	//	scan "\n1D CA file format %d."
	ifs >> dummybuf >> dummybuf >> dummybuf >> dummybuf
		 >> version;

	if (version > 2)
	{  //	 scan "\nCA type %d"
		ifs.get();    // read off eoln
		ifs >> dummybuf >> dummybuf
			 >> itype;
	}
	else
		itype = CA_STANDARD; // FILEVERSION 1 & 2 only have standard

	if (version < 3)
	{  //	 scan "\n%d states and radius %d."
		ifs.get();    // read off eoln
		ifs >> istates >> dummybuf >> dummybuf >> dummybuf >> iradius;
	}

	if (version == 3)	// Al's getting crazy, radius must come first for some reason
	{	//	 scan "\nradius %d  states %d."
		ifs.get();    // read off eoln
		ifs >> dummybuf >> iradius >> dummybuf >> istates >> ch;
	}

	/* We don't write the lookup table to the file if we are dealing
		with a Wave-based CA */
	if (version == 3 && itype >= CA_WAVE)
	{	//	 scan "\nLookup table is not used.\n"
		ifs.get();	// read off eoln
		ifs.getline(dummybuf, sizeof(dummybuf));
	}
	else
	{	//	 scan "\nThe lookup table with %d entries follows:\n"
		ifs.get();	// read off eoln
		ifs.getline(dummybuf, sizeof(dummybuf));
	}
		// just to eat it.  This should have been unsigned int,
		// because 32K shows up as negative.

	target->Settype(itype);
	target->Changeradiusandstates(iradius, istates); //This sets nabeoptions.


	/* We don't write the lookup table to the file if we are dealing
		with a Wave-based CA */
	if (!(version == 3 && target->type_ca >= CA_WAVE))
	{
		j = 0;
		for (i = 0; i < target->nabeoptions; i++)
		{	//	  scan "%u,"
			ifs >> uiread >> ch;
			target->lookup[i] = (unsigned char)uiread;
			if (++j>26)
			{	ifs.get();	// read off eoln
				j = 0;
			}
		}
	}

	if (version == 1)
		target->Setoldcolor();		// FILEVERSTION 1 doesn't save the color table

	if (version == 2)
	{	//	 scan "\nThe color table with Red,Green,Blue values follows:"
		ifs.get();  // read off eoln
		ifs.getline(dummybuf, sizeof(dummybuf));
		j = 0;
		for (i = 0; i < target->states; i++)
		{	//	 scan "%u, %u, %u,"
			ifs >> red >> ch >> green >> ch >> blue >> ch;
			target->colortable[i] = RGB(red,green,blue);
			if (++j>6)
			{	ifs.get();	// read off eoln
				j = 0;
			}
		}
	}

	if (version == 3)
	{	//	 scan "\nThe color table with %d Red,Green,Blue values follows:\n"
		ifs.get();	// read off eoln
		ifs >> dummybuf >> dummybuf >> dummybuf >> dummybuf >> uiread
			 >> dummybuf >> dummybuf >> dummybuf;
		ifs.get();	// read off eoln

		if (uiread > MAX_COLOR + 1)
		{
			MessageBox( target->hwnd, //calist_ptr->hwnd,
				(LPSTR)"Wrong # of colors saved!",
				(LPSTR)"File Problems!",
				MB_OK | MB_ICONEXCLAMATION );
			ifs.close();
			return ifs;
		}

		j = 0;
		for (i = 0; i < target->states; i++)
		{	//	 scan "%u, %u, %u,"
			ifs >> red >> ch >> green >> ch >> blue >> ch;
			target->colortable[i] = RGB(red,green,blue);
			if (++j>6)
			{	ifs.get();	// read off eoln
				j = 0;
			}
		}

		//	scanf "\nShow mode is %"
		ifs.get();  // read off eoln
		ifs >> dummybuf >> dummybuf >> dummybuf >> target->showmode;
		//	scanf "\nGenerator flag is %"
		ifs.get();  // read off eoln
		ifs >> dummybuf >> dummybuf >> dummybuf >> uiread;
		target->generatorflag = (unsigned char) uiread;
		//	scanf "\nWrap flag is %"
		ifs.get();  // read off eoln
		ifs >> dummybuf >> dummybuf >> dummybuf >> uiread;
		target->wrapflag = (unsigned char) uiread;

		// require floats to have decimal point
		ifs >> setiosflags(ios::showpoint);

		//	scanf "\nLambda value of %u.%u"
		ifs.get();  // read off eoln
		ifs >> dummybuf >> dummybuf >> dummybuf >> target->lambda;
		//	scan "\nEntropy level of %u.%u"
		ifs.get();  // read off eoln
		ifs >> dummybuf >> dummybuf >> dummybuf >> target->target_entropy;
		//	scan "\nEntropy score of %u.%u"
		ifs.get();  // read off eoln
		ifs >> dummybuf >> dummybuf >> dummybuf >> target->entropy_bonus;
		//	scan "\nFailstripe level of %u.%u"
		ifs.get();  // read off eoln
		ifs >> dummybuf >> dummybuf >> dummybuf >> target->fail_stripe;

		/* reset all buffers so the loaded information goes into rowbuffer[0] */
		target->sourcerowindex = 0;
		target->targetrowindex = 1;
		target->pastrowindex = MEMORY - 1;
		target->wavesourceindex = 0;
		target->wavetargetindex = 1;
		target->source_row = target->rowbuffer[0];
		target->target_row = target->rowbuffer[1];
		target->past_row = target->rowbuffer[MEMORY - 1];
		target->wave_source_row = target->waverowbuffer[0];
		target->wave_target_row = target->waverowbuffer[1];

		switch (target->type_ca)
		{
			case CA_STANDARD:
				// scan "\nThe source_row with %d values follows:\n"
				ifs.get();
				ifs >> dummybuf >> dummybuf >> dummybuf >> iwidth;
				ifs.getline(dummybuf, sizeof(dummybuf));
				j = 0;
				readwidth =  MIN(iwidth,GetSystemMetrics(SM_CXSCREEN));
				for (i = 0; i < readwidth; i++)
				{
					//	scan "%u,"
					ifs >> uiread >> ch;
					target->source_row[i] = (unsigned char) uiread;
					if (++j>26)
					{	ifs.get();   // read off eoln
						j = 0;
					}
				}
				for (i=readwidth; i<iwidth; i++)
				{
					ifs >> uiread >> ch;
					if (++j>26)
					{	ifs.get();   // read off eoln
						j = 0;
					}
				}
				break;

			case CA_REVERSIBLE:
				// scan "\nThe source_row with %d values follows:\n"
				ifs.get();
				ifs >> dummybuf >> dummybuf >> dummybuf >> iwidth;
				ifs.getline(dummybuf, sizeof(dummybuf));
				j = 0;
				readwidth =  MIN(iwidth,GetSystemMetrics(SM_CXSCREEN));
				for (i = 0; i < readwidth; i++)
				{
					//	 scan "%d,"
					ifs >> uiread >> ch;
					target->source_row[i] = (unsigned char) uiread;
					if (++j>26)
					{	ifs.get();   // read off eoln
						j = 0;
					}
				}
				for (i=readwidth; i<iwidth; i++)
				{
					ifs >> uiread >> ch;
					if (++j>26)
					{	ifs.get();   // read off eoln
						j = 0;
					}
				}

				// scan "\nThe past_row with %d values follows:\n"
				ifs.get();
				ifs >> dummybuf >> dummybuf >> dummybuf >> iwidth;
				ifs.getline(dummybuf, sizeof(dummybuf));
				j = 0;
				for (i = 0; i < readwidth; i++)
				{
					//	scan "%d,"
					ifs >> uiread >> ch;
					target->past_row[i] = (unsigned char) uiread;
					if (++j>26 )
					{	ifs.get();   // read off eoln
						j = 0;
					}
				}
				for (i=readwidth; i<iwidth; i++)
				{
					ifs >> uiread >> ch;
					if (++j>26)
					{	ifs.get();   // read off eoln
						j = 0;
					}
				}
				break;

			case CA_WAVE:
			case CA_HEATWAVE:
			case CA_WAVE2:
			case CA_OSCILLATOR:
			case CA_OSCILLATOR_WAVE:
			case CA_DIVERSE_OSCILLATOR:
			case CA_DIVERSE_OSCILLATOR_WAVE:
			case CA_WATER_WAVE:
				// scan "\nThe wave_source_row with %d values follows:\n"
				ifs.get();
				ifs >> dummybuf >> dummybuf >> dummybuf >> iwidth;
				ifs.getline(dummybuf, sizeof(dummybuf));
				j = 0;
				readwidth =  MIN(iwidth,GetSystemMetrics(SM_CXSCREEN));
				for (i = 0; i < readwidth; i++)
				{
					/* velocity and intensity could be negative */
					//	 scan "%d.%d,"
					ifs >> target->wave_source_row[i].intensity >> ch;
					//	scan "%d.%d,"
					ifs >> target->wave_source_row[i].velocity >> ch;
					if (++j>8)
					{	ifs.get();   // read off eoln
						j = 0;
					}
				}
				for (i=readwidth; i<iwidth; i++)
				{
					ifs >> dummyreal >> ch >> dummyreal >> ch;
					if (++j>8)
					{	ifs.get();   // read off eoln
						j = 0;
					}
				}
				//	 scan "\nTime Step is %u.%u"
				ifs.get();   // read off eoln
				ifs >> dummybuf >> dummybuf >> dummybuf >> target->time_step;
				//	 scan "\nSpace step is %u.%u"
				ifs.get();   // read off eoln
				ifs >> dummybuf >> dummybuf >> dummybuf >> target->space_step;
				//	 scan "\nBand count is %u.%u"
				ifs.get();   // read off eoln
				ifs >> dummybuf >> dummybuf >> dummybuf >> target->band_count;
				//	 scan "\nHeat increment is %u.%u"
				ifs.get();   // read off eoln
				ifs >> dummybuf >> dummybuf >> dummybuf >> target->heat_inc;
				//	 scan "\nMax intensity is %u.%u"
				ifs.get();   // read off eoln
				ifs >> dummybuf >> dummybuf>> dummybuf >> target->max_intensity;
				//	 scan "\nVelocity range is %u.%u"
				ifs.get();   // read off eoln
				ifs >> dummybuf >> dummybuf >> dummybuf >> target->max_velocity;
				//	 scan "\nFriction multiplier is %u.%u"
				ifs.get();   // read off eoln
				ifs >> dummybuf >> dummybuf >> dummybuf >> target->friction_multiplier;
				//	 scan "\nSpring multiplier is %u.%u"
				ifs.get();   // read off eoln
				ifs >> dummybuf >> dummybuf >> dummybuf >> target->spring_multiplier;
				//	 scan "\nDriver multiplier is %u.%u"
				ifs.get();   // read off eoln
				ifs >> dummybuf >> dummybuf>> dummybuf >> target->driver_multiplier;
				//	 scan "\nFrequency multiplier is %u.%u"
				ifs.get();   // read off eoln
				ifs >> dummybuf >> dummybuf>> dummybuf >> target->frequency_multiplier;
				//	 scan "\nMass is %u.%u"
				ifs.get();   // read off eoln
				ifs >> dummybuf >> dummybuf >> target->mass;
				target->Adjust_acceleration_multiplier();
				break;
		}
	}
	target->Computeactual_lambda();
	target->entropy = 0.0;
	target->score = 0;

	return ifs;
}

