#include <windows.h>
#include <stdio.h>
#include <fstream.h>
#include <iomanip.h>
#include "types.h"
#include "bitmap.hpp"
#include "random.h"
#include "camenu.h"

#define CLAMP(x,lo,hi) (x)=(((x)<(lo))?(lo):(((x)>(hi))?(hi):(x)))
/* I was suspicious of the CLAMP macro, so I put a RealWrap form of it
in the ca.cpp.*/

#define MAX_CAS 36 //64 works, but use less for faster comeup.
#define MAXNABEOPTIONS 0x4000U //16K
#define MEMORY 16 /*This must be >= 2.  You need to be careful if it
	ever exceeds horz_count or vert_count.*/
#define MAX_STATES 32
#define MAX_COLOR 255
#define MAX_RADIUS 7
#define HIGH_FITNESS 0.9
#define MAX_LAMBDA 0.90
#define MIN_LAMBDA 0.10
#define BORDER 5
#define FILEVERSION 3
#define WAVE_PERIOD 200
#define GENERATOR_PERIOD 200
#define OSCILLATOR_BASE_PERIOD 200 //This  many timesteps

#define MIN_MULTIPLIER 0.0003 //Used in CA::Changeaccel
#define RAMP_X_MIN 0.1
#define RAMP_X_MAX 0.3
#define RAMP_Y_MIN 0.1
#define RAMP_Y_MAX 0.3
	// used for experiment dialog
//Enums
#define BOTH_SHOW 0
#define ODD_SHOW 1
#define EVEN_SHOW 2
#define ZAP 0	// breed methods
#define CLONE 1
#define SEX 2
#define NONE 3
#define MAX_EVOLVEFLAG 4
#define CA_STANDARD 10
#define CA_REVERSIBLE 11
#define CA_WAVE 12
#define CA_WAVE2 13
#define CA_HEATWAVE 14
#define CA_OSCILLATOR 15
#define CA_OSCILLATOR_WAVE 16
#define CA_DIVERSE_OSCILLATOR 17
#define CA_DIVERSE_OSCILLATOR_WAVE 18
#define CA_WATER_WAVE 19
#define MAX_TYPE_CA 10 //counts number of types.
//Starts
#define START_SHOWMODE BOTH_SHOW
#define START_VIEWMODE IDC_SCROLL_VIEW
#define START_SHOWVELOCITY FALSE
#define START_CAS 9
#define START_WRAPFLAG  0
#define START_ENTROPYFLAG 1
#define START_EVOLVEFLAG ZAP
#define START_MUTATE_FLAG 0
#define START_TARGET_ENTROPY 0.5
#define START_FAILSTRIPE_SCORE -100
#define START_ENTROPY_BONUS 200
#define START_BREED_FLAG 0
#define START_SEED_FLAG	0
#define START_BREED_CYCLE	600
#define START_SEED_CYCLE	400
#define START_SEED_METHOD	1
#define START_LAMBDA 0.3
#define START_MUTATION_STRENGTH 0.10;
#define START_TYPE_CA CA_WAVE
#define START_MIN_VELOCITY 1.0
#define START_MIN_INTENSITY 2.0
#define START_MAX_VELOCITY 100.0
#define START_MAX_INTENSITY 500.0
#define MAX_GENERATOR_INTENSITY 256.0
#define START_HEAT_INC 1.0 //This is actually heat increment.
#define START_BAND_COUNT 64
#define START_MAX_ACCELERATION_MULTIPLIER 0.2
#define START_MIN_MASS 0.1
#define START_MAX_MASS 5.0
#define START_MIN_TIME_STEP 0.0001
#define START_MAX_TIME_STEP 0.5
#define START_MIN_SPACE_STEP 0.0001
#define START_MAX_SPACE_STEP 0.5
#define START_MAX_FRICTION_MULTIPLIER 0.5
#define START_MIN_SPRING 0.1
#define START_MAX_SPRING 5.0
#define START_MAX_DRIVER_MULTIPLIER 1.0
#define START_MAX_FREQUENCY_MULTIPLIER 10.0
#define START_RADIUS 1
#define START_STATES 8
#define START_SCROLL_FLAG 	0 //0 is off, 1 is on
#define START_STRIPEKILL_FLAG 0
#define START_STRIPESEED_FLAG 0
#define START_GENERATOR_FLAG 0
#define START_GENERATOR_LOCATION 0
#define START_WIRE_FLAG 0
#define START_GRAPH_FLAG 0

extern unsigned char timer_flag;	// Take a time reading? 1 - yes
extern unsigned char update_flag; // Update all the dialog boxes? 1 - yes
class CA;
extern ofstream& operator<<(ofstream& ofs, CA* target);
extern ifstream& operator>>(ifstream& ifs, CA* target);

struct Wavecell  //a struct is an all-public class
{
	unsigned char state;
	Real intensity;
	Real velocity;
	Real friction_multiplier;
	Real spring_multiplier;
	Real driver_multiplier;
	Real frequency_multiplier;
	Wavecell():state(0.0), intensity(0.0), velocity(0.0)
	{ //Set these half to 1.0 to 10.0  and half 1.0 to 0.1
  	//multiply them times the CAs corresponding fields in diverseoscillator.
		friction_multiplier = Random(2)?
     	(1.0 + Randomreal()*9.0):
        (1.0/(1.0 + Randomreal()*9.0));
		spring_multiplier =Random(2)?
     	(1.0 + Randomreal()*9.0):
        (1.0/(1.0 + Randomreal()*9.0));
		driver_multiplier =Random(2)?
     	(1.0 + Randomreal()*9.0):
        (1.0/(1.0 + Randomreal()*9.0));
		frequency_multiplier =Random(2)?
     	(1.0 + Randomreal()*9.0):
        (1.0/(1.0 + Randomreal()*9.0));
	} //constructor
};

class Generator
{
	friend class CA;
	friend class CAlist;

private:
	CA *ca_ptr; /* know your owner */
	int location; /* know the x coordinate of your onecell location
		relative to the row buffer of your owner*/
	unsigned char state;
	Real intensity; /* for doing waves */
	Real velocity;
	Real time; /* ticks of time increment in timestep of myca.  At
		present (Rudy November 7, 1994) we are using velocity_
		mulitplier for the timestep. */
public:
	Generator(CA *myca); //constructor.  We use the default destructor.
	void Seed(){intensity = 0.0; time = 0.0; state = 0;}
	void Step();
};

class CA
{
	friend class CAlist;
	friend class Generator;
		  friend ofstream& operator<<(ofstream& ofs, CA* target);
		  friend ifstream& operator>>(ifstream& ifs, CA* target);
private:
//Base
	WindowBitmap *WBM;
	HWND hwnd; //Where I live
	COLORREF colortable[MAX_COLOR+1]; /* for digital CAs, the color
		of a cell will just be colortable[state].  For CA_wave guys,
		we'll do some map from intensity into the MAX_COLOR range and
		then use colortable on that. */
	class CAlist *calist_ptr;	//Know your parent
	unsigned char type_ca; //Know your type.
	int minx, miny, maxx, maxy, horz_count, vert_count; /* These are set
		by Locate according to tile_number and the size of the display */
	int tile_number; //Use this for multiple CAs on one screen
	int computed_lines;
	int row_number; /* a counter to keep track where you are.  This is
		wrapped back to miny after it hits maxy.  Initialized in Locate.*/
	long int cellcount; /*How many cells have you updated so far?  Use
		this with entropy and freqlookup. */
	unsigned char wrapflag; /* 1 means wrap and 0 means line the edges
		with zeroes */
	unsigned char generatorflag;
	class Generator *generator_ptr;
	double fitness; /* lies between 0 and 1, is typically set to
		fabs(entropy - target_entropy). But if a rule fails the
		shift test then its fitness is 0. */
	unsigned char entropyflag; /* 1 means compute entropy every 64K
		cells (and act on it) and 0 means don't */
	double entropy; /* normalized entropy lies between 0.0 and 1.0, it
		tells what percentage of maximum entropy the nabe patterns have.*/
	double target_entropy; /* normalized entropy between 0.0 and 1.0,
		that you want to evolve towards */
	double entropy_bonus;
		/* The score bonus for being close to targ4et_entropy */
	double fail_stripe; /* score penalty for fail strip-ing */
	double score; /* A ca's current score */
	unsigned char sourcerowindex; /* always have source_row =
		rowbuffer[sourcerowindex] */
	unsigned char targetrowindex; /* always have target_row =
		rowbuffer[targetrowindex].  Also keep targetrowindex =
		sourcerowindex + 1 MOD MEMORY.*/
	unsigned char far *source_row;
	unsigned char far *target_row;
		/* use these two buffers to update the CA; the present values are
		read from source_row and the new values are stored in target_row.
		Then you swap these two pointers, show the source_row, and do the
		update again.  These are allocated in Locate, after the actual
		size of the CA is known.*/
	unsigned char far *rowbuffer[MEMORY]; /*Use for remembering old rows and
		checking for repeats stripes.*/
	unsigned int far *freqlookup;
		/* freqlookup keeps track of how often each nabeoption occurs.
		Whenever you check lookup[nabecode], you at the same time
		increment freqlookup[nabecode].  Note that you can run ths over
		64K really quickly, like in 100 lines of fullscreen.  But suppose
		we only use in screens of 4 or 9 CAs.  In 4, it takes 200, not
		quite a full frame.  Well, that'll do.  And I'll check entropy
		each time the worst case count hits 64K.  unsigned char would
		be out of the question, like one line even in 9 mode.  Would
		prefer long unsigned int, kind of, but that's overkill.  At
		the worst, counting lookup and freqtalbe, we are looking at
		9*3*32K  = 864K bytes, which won't work.  So there has to be
		a check on it.*/
//Digital
	int radius; //radius of cells to examine, range 1 to 4.
	int states; //number of states per cell. Only 2 is allowed right now.
	int statebits; //number of bits per state.
	int nabesize; // 1 + 2*radius, the number of cells in a neighborhood.
	unsigned short nabeoptions; //Keep this under 32K
		/* pow(2,nabesize), the number of possible lookup
		tables.  For radius 1,2,3,4, this is 8,32,128,512. */
	unsigned short mask;
		// nabeoptions - 1. Use to mask the nabecode in Update.
	double lambda;
		/* lambda measures the density of non-zero entries desried in the
		lookup table.  Start with START_LAMBDA.*/
	double actual_lambda;
		/* actuallambda measures the density of non-zero entries that are
		actually in the lookup table.  This needs to be kept separate
		from the *desired* value lambda because the value drifts in
		a brownian walk, and can end up stuck at 0.0 or 1.0 unless
		you keep normalizing it to the desired value lambda.*/
	unsigned char far *lookup;
		/* lookup holds the new cell values, indexed by the nabe
		code numbers which are assembled in Update.  lookup is allocated
		in the constructor.  Lookups are all well under 32K,
		so don't need huge.*/
//Reversible private
	unsigned char pastrowindex;
	unsigned char far *past_row;
//Wave private
	int band_count;
	float max_intensity;
	float max_velocity;
	int showmode;
	Real time; /* We are using velocity_mulitplier for the timestep. */
	Real time_step, space_step, acceleration_multiplier, half_time_step;
	/* We update newintensity += time_step * velocity and
		newvelocity += acceleration_multiplier * delta_intensity,
		where delta_intensity is, in the one-neighbor case,
		-2C_intensity+L_intensity+R_intensity.
		acceleration_multiplier is a function of  space_step,
		and time_step, which is set by Adjust_acceleration_multiplier().*/
	Wavecell far *wave_source_row;
	Wavecell far *wave_target_row;
	unsigned char wavesourceindex; /* always have wave_source_row =
		waverowbuffer[wavesourceindex] */
	unsigned char wavetargetindex; /* always have wave_target_row =
		waverowbuffer[wavetargetindex] */
	Wavecell far *waverowbuffer[2];
//Wave otiose?
	int oldstates;
		// used to restore states & radius going back to standard
	int oldradius;
//Oscillator private
	Real friction_multiplier, spring_multiplier, driver_multiplier,
		frequency_multiplier, mass, frequency_factor;//frequency_factor
     //is a helping variable defined in :Adjust_acceleration as
     // 2*PI/(time_step*OSCILLATOR_BASE_PERIOD), so that BASE_PERIOD
     // steps puts it thru a full oscillation.
//Heat private
	Real heat_inc; /* This is the amount to increment the intensity if
		you are number of steps to wait if you are doing driven heat */
public:
//Base public
	CA(CAlist *mylist);
		/* Sets Allocates. constructor is called when a CA object is
		 declared.  I have the constructor set all the possible fields.
		adopts the hwnd of mylist, adopts mylist as calist_ptr.*/
	~CA(); /* The destructor.  Frees the memory in source_row,
		target_row, lookup, etc.*/
	void Allocate();
		/*Allocates rowbuffer[i], lookup and freqlookup  handles*/
	int Gettype(void) {return type_ca;}
	void Settype(int newtype);
	void Locate(int itile_number = 0,int dmaxx = 0, int dmaxy = 0,
		int CA_count_per_edge = 1);
		/*tile_count tells how many CAs you want on screen, and
		itile_number tells which position this CA is in.  itile_number
		lies between 0 and tile_count-1.  Locate sets minx,maxx,miny,
		maxy, and intializes row_number to miny.Locate sets horz_count to
		maxx + 1 - minx, which is the and allocates source_row and
		target_row.*/
	void InitSeed();
		/* This sets all cells in all rows to zero. This is called
		initially from the constructor */
	void Seed();
		/* this seeds the source_row with random numbers
		between 0 and states - 1.  It is called in Locate at startup,
		after source_row is allocated.  Let each Wavecell.intensity and
		Wavecell.velocity range from -MAX_INTENSITY to +MAX_INTENSITY,
		and let the velocity be 0.0*/
	void Sineseed();
		/* This seeds the source row with a periodic number pattern.
		Used especially for waves, where it seeds intentsity as a sine
		wave and velocity as the corresponding derivative cosine wave.*/
	void Oneseed();
		/* this seeds the source_row with all zeroes except for a
		single one in the center.  Set all the Wavecell's intensity
		and velocity to 0.0, except the central one, which gets a
		random intensity between -MAX_INTENSITY to +MAX_INTENSITY*/
	void Zeroseed();
		/* seeds with all zeros to calm down so you can see the
		generator*/
	void SyncRows();
		/* Copy source_row[0 to horz_count] into rowbuffer[0],
		copy past_row[0 to horz_count] into rowbuffer[MEMORY - 1],
		copy wave_source_row[0 to horz_count] back to start_wave_row
		this only does something if source and target are flipped
		from their original positions.
		This is because when we seed in the beginning, we only seed
		rowbuffer[0] with the 'whole' line.
		We need to do this when we resize, because we need the extra
		values stored in rowbuffer[0]. File save also uses this because
		we want to save the whole line */
	int Getentropyflag(void) {return entropyflag;}
	void Setentropyflag(int onoff) {entropyflag = onoff;
		 if(entropyflag)Resetfreq();}
	double Gettargetentropy(void) {return target_entropy;}
		// desired entropy
	double Getentropy(void) {return entropy;}	// actual entropy
	unsigned char Getwrapflag(void) {return wrapflag;}
	void Setwrapflag(int onoff) {wrapflag = onoff;}
	unsigned char Getgeneratorflag(void) {return generatorflag;}
	void Setgeneratorflag(int onoff);
	double Getfailstripe(void) {return fail_stripe;}
	void Setfailstripe(int updown);
		// Penalty for failing the stripe test
	double Getscore(){return score;}
	double Getentropybonus(void) {return entropy_bonus;}
	void Setentropybonus(int updown);
		// Bonus for getting an entropy close to target_entropy
	void Resetscore() {score = 0;}	// CA's score
	void Prepareforentropy();
		/* This gets freqlookup ready according as to whether
		entropyflag is 0 for off or 1 for on */
	void CopyCA(CA *goodCA);
		/* Copy everything from goodCA. Wave stuff, source_row, ... */
	void Mutate(float mutation_strength);
		/* mutation_strength between 0.0 and 1.0.  Mutate that
		proportion of the table's entries*/
	void Setcolortable();
		/* Depending on the number of states, we set color values.*/
	void Setoldcolor();
		/* Try to get the good old 16 values of Rudy's color table */
	void Changeentropy(double newentropy);
		/* increment or decrement the target_entropy */
	void Become_child_of(CA *mom, CA *dad);
		/* Adopt mom's states and radius.  Make your lookup be
		a combination of mom's and dad's.  Use the part of mom's
		and part of dad's.  If dad's is shorter, copy his whole into
		the start of mom's lookup.*/
	void Entropy();
		/* Use freqlookup to compute the entropy if the entropyflag is
		on.  This is called everytime that cellcount passes 64K.  The
		fitness is set according to the entropy compared to
		the target_entropy.*/
	void Resetfreq();
		/* set cellcount to zero and zero out the freqlookup table
		if entropyflag is on*/
	void Avoidstripes(); // If the CA has a reoccuring pattern, Seed it.
	void RealWrap(Real &vx, Real vlo, Real vhi);
	void Install_Ramp(int paramcode, double rampvalue);
///Digital public
	int Getradius(void) {return radius;}
	int Getstates(void) {return states;}
	unsigned short Getnabe(void) {return nabeoptions;}
		// Returns the number of nabeoptions. Used by Lookup Dialog Box
	COLORREF Color_look(int numb);
		// Returns colortable[lookup[numb]]. Used by Lookup Dialog Box???
	float Getlambda(void) {return lambda;}
	float Getactuallambda(void) {return actual_lambda;}
	void Lambdalookup();
		// Sets a new random lookup with lambda percent non-zero values.
	void Changeradiusandstates(int newradius = -1, int newstates = -1);
	// if newradius or newstates == -1, then that part is unchanged
		/* These reset radius, states, nabesize, nabeoptions, and
		mask, frees the old lookup, and allocates a new one and
		fills it.  We need the combined form for copying CAs so that
		a legit copy isn't rejected because e.g. the radius looks too
		big before you notice the new state makes it ok. */
	void Symmetrize(); /* Make the lookup table summetric */
	void Setlambda(float newlambda);
	void Changelambda(float newlambda);
		/* Increment or decrement lambda if updown is, respectively
		1 or -1.  In the range 0.1 to 0.9, change by adding or
		subtracting 0.1.  Out of that range chagne by halving the
		distance to, respectively, 1.0 or 0.0.  Then change the current
		lookup table's density of nonzero entries to the new lambda.*/
	void Computeactual_lambda();	// Find a CAs actual lambda level
//Reversible public
	void Reverse(); /* swap the past_row entries with the
		source_row entries.  This makes a TYPE_CA_REVERSIBLE
		rule run backwards.  It works like this.  In a rule
		like this,
		Target = f(Source) - Past, so also
		Past = f(Source) - Target.  At the end of a call to
		ReversibleUpdate the buffers get rolled forward.  So
		if I start with Past0 and Source0 and compute Target0,
		then Past1 is Source0 and Source1 is Target0.  If I
		exchange the entries in these two to get Source2 =
		Past1 and Past2 = Source1, then when I update
		Target1 = f(Source2) - Past2, I am really doing
		Target1 = f(Past1) - Source1, which is
		Target1 = f(Source0) - Target0, which  is Past0.*/
//Wave public
	float Getmaxintensity(void) {return max_intensity;}
	float Getmaxvelocity(void) {return max_velocity;}
	void Change_acceleration_multiplier(int updown);
	void Change_time_step(int updown);
	void Change_space_step(int updown);
	void Adjust_acceleration_multiplier();
	void Change_band_count(int updown);
	void Change_max_intensity(int updown);
	void Change_max_velocity(int updown);
	void Change_spring(int updown);
	void Change_mass(int updown);
	void Change_friction(int updown);
	void Change_driver_amp(int updown);
	void Change_driver_freq(int updown);
	Real Get_spring(void) {return spring_multiplier;}
	Real Get_mass(void){return mass;}
	Real Get_friction(void) {return friction_multiplier;}
	Real Get_driver_amp(void) {return driver_multiplier;}
	Real Get_driver_freq(void) {return frequency_multiplier;}
	Real Get_space_step(void) {return space_step;}
	Real Get_time_step(void) {return time_step;}
	int Getband(void) {return band_count;}
	int Getshowmode(){return showmode;}
	void Setshowmode(int newmode){showmode = newmode;}
	void Set_time_step(double ts);
	void Set_space_step(double ss);
	void Set_band_count(double bc);
	void Set_max_velocity(double mv);
	void Set_max_intensity(double mi);
	void Set_heat_inc(double hi);
	void Set_spring(double s);
	void Set_friction(double f);
	void Set_driver_amp(double da);
	void Set_driver_freq(double df);
	void Set_mass(double m);
//Heat public
	Real Get_heat_inc(){return heat_inc;}
	void Change_heat_inc(int updown);
//Update functions
	void (CA::*UpdateFunction)(HDC hdc);  /* This is a member function
		pointer as decribed to me by Horstmann, with a reference to
		 "Coplein's book on advanced C++ idioms".*/
	void Update(HDC hdc); //Define in ca.cpp as a switch on ca_type.
	void (CA::*UpdateCell)(int, int, int);
		/* This is a member function pointer used by Wave functions
		to update the velocity and intinsity*/
	void (CA::*UpdateCell2)(int, int, int, int, int);
		/* This is a member function pointer used by Wave functions
		to update the velocity and intinsity*/
/*	void (CA::*BoundIntFunction)(Real &r);
	void (CA::*BoundVelFunction)(Real &r);
	void ClampBound(Real &r);
	void WrapBound(Real &r);
	void KillBound(Real &r);
*/
	void Show(HDC hdc); /* This holds the code common to the STandard
		Updates and the Wave Updates, the show & stripe stuff, also
		the stuff dealing with wire mode. */
	void StandardUpdate(HDC hdc);
		/* this is the heart of the algorithm.  You update the values
		of target_row by successively converting each cell's neighborhood
		into a bit-assembled nabe code number.  When states is 2, you use
		repeated calls of nabe <<= 1, nabe&= mask, and
		nabe |= source_row[i], and then compute target_row[i] as
		lookup[nabe].  When target_row is ready, you swap its pointer
		with source_row.  And then you increment row_number.*/
	void ReversibleUpdate(HDC hdc);
		/* This is the same as the CA_standard Update except that
		while before, you have NewC = lookup(C_nabe), you now let
		NewC = ( f(C_nabe) - OldC ) MOD (states).*/
	void WaveUpdate(HDC hdc);
		/* The wave update method is to compute acceleration as
		(2* C_intensity) - (L_intensity + R_intensity),
		to then do velocity += acceleration_multiplier * acceleration,
		and then do intensity += time_step * velocity.*/
	void WaveUpdate2(HDC hdc);
		/* Two neighbors either side.*/
	void WaveUpdateStep(HDC hdc);
		/* This holds the code common to WaveUpdate and WaveUpdate2, all
		the stepping stuff.*/
	void WaveVelInt1(int, int, int);
	void WaveVelInt2(int, int, int, int, int); // five nabe
	void HeatwaveUpdate(HDC hdc);
		/* does Wave with an averaging every SMOOTHCOUNT */
	void HeatInt1(int, int, int);
	void HeatInt2(int, int, int, int, int);
	void Oscillator(int, int, int);
	void OscillatorWave(int, int, int);
	void DiverseOscillator(int, int, int);
	void DiverseOscillatorWave(int, int, int);
	void WaterWave(int, int, int, int, int);
	void UlamWave(int, int, int);
};

class CAlist
{
	friend class CA;
private:
//Base private
	HWND hwnd; //Where I live
	WindowBitmap *WBM; // Bitmap. Put pixels here, then bitblt all to screen
	int count; // Number of CA objects in the list.
	CA *list[MAX_CAS]; // Pointer list to the CAs
	CA *focus; //This is the CA with the box around it.
	unsigned char zoomflag; //If this is on, we only show the focus CA.
	float mutation_strength; // Amount to mutate 0 to 1.
	int CA_count_per_edge; // length & height of displayed cells
	int sleep; /* Used when the window is minimized. If sleep = 1,
		then update and show will not work. */
	int best_list[MAX_CAS]; /* This has all the CA position numbers sorted
		out by the best score. Set by Sortfittness() */
	unsigned char mutateflag; // Mutate all CAs at breed cycle time?
	unsigned char breedflag; 	// Enable breed cycle?
	unsigned int breedcycle; // We breed after this many cycles
	int breedcycle_count; // breed cycle counter
	unsigned char evolveflag; /* Methods to breed. 0, 1, 2 are
		Zap, Clone, and Sex. */
	int viewmode; //IDC_DOWN_VIEW, IDC_SCROLL_VIEW,
		//IDC_WIRE_VIEW or IDC_GRAPH_VIEW.
	int showvelocity; //0 means show intensity, 1 means showvelocity
	unsigned char graphflag; // if 1, then a graph view
	unsigned char wireflag;  // if 1, then a wire shape CA is displayed
	unsigned char scrollflag;	// If 1, then we scroll the CA's else wrap to top
	unsigned char stripekillflag; /* If 1, then we do a lamndalookup right away on
		somebody who fails the stripe test, otherwise we add the penalty to
		his score */
	unsigned char stripeseedflag; //seed if stripe fail
public:
//Base public
	CAlist(HWND myhwnd, int icount); //constructor
	~CAlist(); //destructor
	int Count(){return count;} //listlength
	CA *FocusCA(){return focus;}	// This is a pointer to the focus CA
	CA *GetCA(int i) {return list[i];}  // returns ith CA
	void Locate(int dmaxx, int dmaxy); /*This gnarly method uses count and
		dp to figure out where to tell the list[i] CAs to put their
		minx, maxx, miny, maxy.*/
	int Getbreedingsteps() {return breedcycle;} // Breed after this many cycles
	int Changecount(int icount); /* Used by expand and shrink. Changes the
		number of CAs shown on the screen to icount */
	int Bumpcount(int updown); /* Adds or takes away a whole CA_count_per_edge of
		CAs. e.g. 3x3 to 4x4 or vise versa. Updown is 1 or -1 */
	int Zoom(int inout); /* 1 for in (1 CA), -1 for out (the active CAs).
		When you do Zoom(1), you move focus to the CA you get to see.
		Return a 0 if you're repeating a zoom.*/
	void Boxfocus(HDC hdc, COLORREF color); //Draw a box around focus CA.
	void Seed(); //Seed all active CAs.
	void Update_and_Show(HDC hdc);  /*Update all active CAs and show the
		updates in a rapid way.*/
	void Setentropy(int onoff); /* 0 for off, 1 for on */
	void Setwrapflag(int onoff);	/* 0 for off, 1 for on */
	void Setgeneratorflag(int onoff);	/* 0 for off, 1 for on */
	void Setscrollflag(int onoff) { scrollflag = onoff; }
	int Getviewmode(){return viewmode;}
	int Getshowvelocity(){return showvelocity;}
	void Setshowvelocity(int newshowvelocity){showvelocity = newshowvelocity;}
   void Setviewmode(int newmode);
	unsigned char Getscrollflag(void) { return scrollflag; }
	unsigned char Getwireflag(void) { return wireflag; }
	unsigned char Getgraphflag(void) { return graphflag; }
	void Setgraphflag(unsigned int onoff) { graphflag = onoff; }
	void Setwireflag(unsigned int onoff) { wireflag = onoff; }

	void Setstripekillflag(int onoff) { stripekillflag = onoff; }
	unsigned char Getstripekillflag(void) { return stripekillflag; }
	void Setstripeseedflag(int onoff) { stripeseedflag = onoff; }
	unsigned char Getstripeseedflag(void) { return stripeseedflag; }
   void DrawDivider(HDC hdc);
	CA *Getfocus(int cursorx, int cursory);	/* return pointer to the CA
		that is positioned on screen at cursorx,cursory */
	int Setfocus(HDC hdc, CA *newfocus);	/* Change focus CA */
	void Evolve();/* After each breed cycle this is called
		depending what evolveflag is set to this can happen:
		0-zap - mutates everybody
		1-clone - clones the best ones and mutates all
		2-breed - breeds the best ones and mutates all */
	void Sortfitness(); /* This uses the best_list[] array to sort
		all the CAs by their score.  The best_list[] array holds
		CA numbers */
	float Mutation_strength(){return mutation_strength;} /* How much to
		mutate 0 to 1.0 */
	void Changeentropy(double newentropy);	// Changes the target_entropy
	void Setcolortable();	// Call each CA's setcolortable
	void Oneseed();	// Call each CA's oneseed
	void Zeroseed(); //Call each CA's zeroseed
	void Sineseed();	// Call each CA's sineseed
	void Mutate();	// Mutate each CA by mutation_strength
	void Breed();	// Makes all the CAs a child of the focus and top left
	void Copymutate();	// Copys a mutation of the focus to each CA
	void SetWindowBitmap(WindowBitmap *myWBM);	/* Initializes the Bitmap
		pointer for use by calist and each CA gets a copy also*/
	void SetSleep(int onoff); // 1 - Window is minimized, dont update or show
	int GetSleep(){return sleep;} // Use with IDM_PAUSE in capow.cpp
	void ToggleSleep(){sleep ^= 1;}//Ditto
	void Setmutation(int updown);
	float Getmutation() {return mutation_strength;}
		/* Mutation strength is a percentage 0.0 to 1.0 of how much
		of the lookup you wanted mutated */
	unsigned char Getmutateflag() {return mutateflag;}
		// Mutate at Breed cycle?
	void Setmutateflag(unsigned char imutate) {mutateflag = imutate;}
	unsigned char Getbreedflag() {return breedflag;}
		// Breed cycle enabled?
	void Setbreedflag(unsigned char ibreed) {breedflag = ibreed;}
	void Setfailstripe(int updown); // fail stripe penalty
	void Setentropybonus(int updown); // bonus for a close entropy
	void Setbreedcycle(int updown); // When do we Breed? 1 or -1
	int Getevolve(void) {return evolveflag;}
		/* Sets and Gets the ca's evolve flag for use by Cycle
		0 - Zap, 1 - Clone, 2 - Breed */
	void SetEvolve(unsigned char newevolve) {evolveflag = newevolve;}
	void Resetscores();	// Sets all CA scores to 0
	void SetCAType(CA *target, unsigned char newtype);
		/* Changes the type of the CA*/
	void SetAllType(unsigned char newtype);
		/* Changes the type of all the cas	*/
	void CopyCA(CA *ca1, CA *ca2);
		/* Copies ca1 to ca2 */
	void Randomize(); //Randomizes type_ca, state, radius
	void Install_Ramp(int paramcode, double *rampvalue);
	void CheckRange (int paramcode, double minmax[]);
		  void Saveindividual(char *filename, CA *target); // Save focus CA
		  void Saveall(char *filename); // Save entire experiment
		  void Loadindividual(char* filename, CA *target); // Load on focus CA
		  void Loadall(char* filename); // Load entire experiment
///Digital public
	void Changeradiusandstates(int newradius = -1, int newstates = -1);
	// if newradius or newstates == -1, then that part is unchanged
	void Lambdalookup(); // Call each CA's lambda lookup.
	void Symmetrize(); // Call each CA's symmetrize
	void Changelambda(float newlambda); // Change all CA's lambda
//Reversible public
	void Reverse(); // Call each CA's Reverse
//Wave public
	void Change_acceleration_multiplier(int updown);
	void Change_time_step(int updown);
	void Change_space_step(int updown);
	void Change_band_count(int updown);
	void Change_max_intensity(int updown);
	void Change_max_velocity(int updown);
	void Change_spring(int updown);
	void Change_mass(int updown);
	void Change_friction(int updown);
	void Change_driver_amp(int updown);
	void Change_driver_freq(int updown);
	void Setshowmode(int newmode);
//Heat public
	void Change_heat_inc(int updown);
};
